# pylint: disable=too-many-lines
"""
CLI Plugin for handling OBJ
"""
import argparse
import getpass
import glob
import math
import os
import platform
import socket
import sys
import time
from datetime import datetime
from math import ceil

from terminaltables import SingleTable

ENV_ACCESS_KEY_NAME = "LINODE_CLI_OBJ_ACCESS_KEY"
ENV_SECRET_KEY_NAME = "LINODE_CLI_OBJ_SECRET_KEY"

try:
    import boto
    from boto.exception import BotoClientError, S3CreateError, S3ResponseError
    from boto.s3.connection import OrdinaryCallingFormat
    from boto.s3.key import Key
    from boto.s3.prefix import Prefix

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False


# replace {} with the cluster name
BASE_URL_TEMPLATE = "{}.linodeobjects.com"
BASE_WEBSITE_TEMPLATE = "website-{}.linodeobjects.com"

# for all date output
DATE_FORMAT = "%Y-%m-%d %H:%M"
INCOMING_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

# for help commands
PLUGIN_BASE = "linode-cli obj"

# constant error messages
NO_SCOPES_ERROR = """Your OAuth token isn't authorized to create Object Storage keys.
To fix this, please generate a new token at this URL:

  https://cloud.linode.com/profile/tokens

Be sure to select 'Read/Write' for Object Storage and 'Read Only'
for Account during token generation (as well as any other access
you want the Linode CLI to have).

Once you've generated a new token, you can use it with the
Linode CLI by running this command and entering it:

  linode-cli configure
"""

NO_ACCESS_ERROR = """You are not authorized to use Object Storage at this time.
Please contact your Linode Account administrator to request
access, or ask them to generate Object Storage Keys for you."""


# Files larger than this need to be uploaded via a multipart upload
UPLOAD_MAX_FILE_SIZE = 1024 * 1024 * 1024 * 5
# This is how big (in MB) the chunks of the file that we upload will be
MULTIPART_UPLOAD_CHUNK_SIZE_DEFAULT = 1024


def restricted_int_arg_type(max, min=1):  # pylint: disable=redefined-builtin
    """
    An ArgumentParser arg type for integers that restricts the value to between `min` and `max`
    (inclusive for both.)
    """

    def restricted_int(string):
        err_msg = f"Value must be an integer between {min} and {max}"
        try:
            value = int(string)
        except ValueError as e:
            # argparse can handle ValueErrors, but shows an unfriendly "invalid restricted_int
            # value: '0.1'" message, so catch and raise with a better message.
            raise argparse.ArgumentTypeError(err_msg) from e
        if value < min or value > max:
            raise argparse.ArgumentTypeError(err_msg)
        return value

    return restricted_int


def list_objects_or_buckets(get_client, args):
    """
    Lists buckets or objects
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " ls")

    parser.add_argument(
        "bucket",
        metavar="NAME",
        type=str,
        nargs="?",
        help="Optional.  If not given, lists all buckets.  If given, "
        "lists the contents of the given bucket.  May contain a "
        "/ followed by a directory path to show the contents of "
        "a directory within the named bucket.",
    )

    parsed = parser.parse_args(args)
    client = get_client()

    if parsed.bucket:
        # list objects
        if "/" in parsed.bucket:
            bucket_name, prefix = parsed.bucket.split("/", 1)
        else:
            bucket_name = parsed.bucket
            prefix = ""

        try:
            bucket = client.get_bucket(bucket_name)
        except S3ResponseError:
            print("No bucket named " + bucket_name)
            sys.exit(2)

        data = []
        for c in bucket.list(prefix=prefix, delimiter="/"):
            if isinstance(c, Prefix):
                size = "DIR"
            else:
                size = c.size

            # pylint: disable-next=redefined-outer-name
            datetime = _convert_datetime(c.last_modified) if size != "DIR" else " " * 16

            data.append([datetime, size, c.name])

        if data:
            tab = _borderless_table(data)
            print(tab.table)

        sys.exit(0)
    else:
        # list buckets
        buckets = client.get_all_buckets()

        data = [[_convert_datetime(b.creation_date), b.name] for b in buckets]

        tab = _borderless_table(data)
        print(tab.table)

        sys.exit(0)


def create_bucket(get_client, args):
    """
    Creates a new bucket
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " mb")

    parser.add_argument(
        "name", metavar="NAME", type=str, help="The name of the bucket to create."
    )

    parsed = parser.parse_args(args)
    client = get_client()

    client.create_bucket(parsed.name)

    print(f"Bucket {parsed.name} created")
    sys.exit(0)


def delete_bucket(get_client, args):
    """
    Deletes a bucket
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " rb")

    parser.add_argument(
        "name", metavar="NAME", type=str, help="The name of the bucket to remove."
    )
    parser.add_argument(
        "--recursive",
        action="store_true",
        help="If given, force removal of non-empty buckets by deleting "
        "all objects in the bucket before deleting the bucket.  For "
        "large buckets, this may take a while.",
    )

    parsed = parser.parse_args(args)
    client = get_client()

    if parsed.recursive:
        try:
            bucket = client.get_bucket(parsed.name)
        except S3ResponseError:
            print("No bucket named " + parsed.name)
            sys.exit(2)

        for c in bucket.list():
            print(f"delete: {parsed.name} {c.key}")
            bucket.delete_key(c)

    client.delete_bucket(parsed.name)

    print("Bucket {parsed.name} removed")

    sys.exit(0)


def upload_object(get_client, args):  # pylint: disable=too-many-locals
    """
    Uploads an object to object storage
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " put")

    parser.add_argument(
        "file", metavar="FILE", type=str, nargs="+", help="The files to upload."
    )
    parser.add_argument(
        "bucket", metavar="BUCKET", type=str, help="The bucket to put a file in."
    )
    parser.add_argument(
        "--acl-public",
        action="store_true",
        help="If set, the new object can be downloaded without " "authentication.",
    )
    parser.add_argument(
        "--chunk-size",
        type=restricted_int_arg_type(5120),
        default=MULTIPART_UPLOAD_CHUNK_SIZE_DEFAULT,
        help="The size of file chunks when uploading large files, in MB.",
    )
    # parser.add_argument('--recursive', action='store_true',
    #                    help="If set, upload directories recursively.")

    parsed = parser.parse_args(args)
    client = get_client()

    to_upload = []
    to_multipart_upload = []
    for c in parsed.file:
        # find the object
        file_path = os.path.expanduser(c)

        # Windows doesn't natively expand globs, so we should implement it here
        if platform.system() == "Windows" and "*" in file_path:
            results = glob.glob(file_path, recursive=True)
            if len(results) < 1:
                print(f"No file found matching pattern {file_path}")
                sys.exit(5)

            if len(results) > 1:
                print(
                    "warn: Found multiple files matching pattern "
                    f"{file_path}, using {results[0]}"
                )

            file_path = results[0]

        if not os.path.isfile(file_path):
            print("No file {file_path}")
            sys.exit(5)

        filename = os.path.split(file_path)[-1]

        file_size = os.path.getsize(file_path)

        if file_size >= UPLOAD_MAX_FILE_SIZE:
            to_multipart_upload.append((filename, file_path, file_size))
        else:
            to_upload.append((filename, file_path))

    # upload the files
    try:
        bucket = client.get_bucket(parsed.bucket)
    except S3ResponseError:
        print("No bucket named " + parsed.bucket)
        sys.exit(2)

    policy = "public-read" if parsed.acl_public else None
    chunk_size = 1024 * 1024 * parsed.chunk_size

    for filename, file_path in to_upload:
        k = Key(bucket)
        k.key = filename

        print(filename)
        k.set_contents_from_filename(file_path, cb=_progress, num_cb=100, policy=policy)

    for filename, file_path, file_size in to_multipart_upload:
        _do_multipart_upload(bucket, filename, file_path, file_size, policy, chunk_size)

    print("Done.")


def _do_multipart_upload(
    bucket, filename, file_path, file_size, policy, chunk_size
):  # pylint: disable=too-many-arguments
    """
    Handles the internals of a multipart upload for a large file.

    :param bucket: The bucket to upload the large file to
    :type bucket: Boto bucket
    :param filename: The name of the file to upload
    :type filename: str
    :param: file_path: That absolute path to the file we're uploading
    :type file_path: str
    :param file_size: The size of this file in bytes (used for chunking)
    :type file_size: int
    :param policy: The canned ACLs to include with the new key once the upload
                   completes.  None for no ACLs, or "public-read" to make the
                   key accessible publicly.
    :type policy: str
    :param chunk_size: The size of chunks to upload, in bytes.
    :type chunk_size: int
    """
    upload = bucket.initiate_multipart_upload(filename, policy=policy)

    # convert chunk_size to float so that division works like we want
    num_chunks = int(math.ceil(file_size / float(chunk_size)))

    print(f"{filename} ({num_chunks} parts)")

    num_tries = 3
    retry_delay = 2

    try:
        with open(file_path, "rb") as f:
            for i in range(num_chunks):
                print(f" Part {i + 1}")
                for attempt in range(num_tries):
                    try:
                        upload.upload_part_from_file(
                            f, i + 1, cb=_progress, num_cb=100, size=chunk_size
                        )
                    except S3ResponseError:
                        if attempt < num_tries - 1:
                            print(
                                f"  Part failed ({attempt+1} of {num_tries} attempts). "
                                f"Retrying in {retry_delay} seconds..."
                            )
                            time.sleep(retry_delay)
                            continue
                        raise
                    else:
                        break
    except Exception:
        print("Upload failed!  Cleaning up!")
        upload.cancel_upload()
        raise

    upload.complete_upload()


def get_object(get_client, args):
    """
    Retrieves an uploaded object and writes it to a file
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " get")

    parser.add_argument(
        "bucket", metavar="BUCKET", type=str, help="The bucket the file is in."
    )
    parser.add_argument(
        "file", metavar="OBJECT", type=str, help="The object to retrieve."
    )
    parser.add_argument(
        "destination",
        metavar="LOCAL_FILE",
        type=str,
        nargs="?",
        help="The destination file. If omitted, uses the object "
        "name and saves to the current directory.",
    )

    parsed = parser.parse_args(args)
    client = get_client()

    # find destination file
    destination = parsed.destination

    if destination is None:
        destination = parsed.file

    # download the file
    try:
        bucket = client.get_bucket(parsed.bucket)
    except S3ResponseError:
        print("No bucket named " + parsed.bucket)
        sys.exit(2)

    k = bucket.get_key(parsed.file)

    if k is None:
        print(f"No {parsed.file} in {parsed.bucket}")
        sys.exit(2)

    k.get_contents_to_filename(destination, cb=_progress, num_cb=100)

    print("Done.")


def delete_object(get_client, args):
    """
    Removes a file from a bucket
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " del")

    parser.add_argument(
        "bucket", metavar="BUCKET", type=str, help="The bucket to delete from."
    )
    parser.add_argument(
        "file", metavar="OBJECT", type=str, help="The object to remove."
    )

    parsed = parser.parse_args(args)
    client = get_client()

    # get the key to delete
    try:
        bucket = client.get_bucket(parsed.bucket)
    except S3ResponseError:
        print("No bucket named " + parsed.bucket)
        sys.exit(2)

    k = bucket.get_key(parsed.file)

    if k is None:
        print(f"No {parsed.file} in {parsed.bucket}")
        sys.exit(2)

    # delete the key
    k.delete()

    print(f"{parsed.file} removed from {parsed.bucket}")


def generate_url(get_client, args):
    """
    Generates a URL to an object
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " signurl")

    parser.add_argument(
        "bucket", metavar="BUCKET", type=str, help="The bucket containing the object."
    )
    parser.add_argument(
        "file", metavar="OBJECT", type=str, help="The object to sign a URL to."
    )
    parser.add_argument(
        "expiry",
        metavar="EXPIRY",
        type=str,
        help="When this link should expire.  Treated as an epoch "
        "time if a number. If starts with a '+' treated as "
        "an offset.",
    )

    parsed = parser.parse_args(args)
    client = get_client()

    now = datetime.now()

    if parsed.expiry.startswith("+"):
        # this is an offset in seconds
        offset = int(parsed.expiry[1:])
    else:
        expiry = int(parsed.expiry)
        offset = expiry - ceil(now.timestamp())

    # get the key
    try:
        bucket = client.get_bucket(parsed.bucket)
    except S3ResponseError:
        print("No bucket named " + parsed.bucket)
        sys.exit(2)

    k = bucket.get_key(parsed.file)

    if k is None:
        print(f"No {parsed.file} in {parsed.bucket}")
        sys.exit(2)

    url = k.generate_url(offset)

    print(url)


def set_acl(get_client, args):
    """
    Modify Access Control List for a Bucket or Objects
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " setacl")

    parser.add_argument(
        "bucket", metavar="BUCKET", type=str, help="The bucket to modify."
    )
    parser.add_argument(
        "file",
        metavar="OBJECT",
        type=str,
        nargs="?",
        help="Optional.  The object to modify.  If omitted, modifies "
        "the ACLs for the entire bucket.",
    )
    parser.add_argument(
        "--acl-public",
        action="store_true",
        help="If given, makes the target publicly readable.",
    )
    parser.add_argument(
        "--acl-private", action="store_true", help="If given, makes the target private."
    )

    parsed = parser.parse_args(args)
    client = get_client()

    # make sure the call is sane
    if parsed.acl_public and parsed.acl_private:
        print("You may not set the ACL to public and private in the same call")
        sys.exit(1)

    if not parsed.acl_public and not parsed.acl_private:
        print("You must choose an ACL to apply")
        sys.exit(1)

    # get the bucket
    try:
        bucket = client.get_bucket(parsed.bucket)
    except S3ResponseError:
        print("No bucket named " + parsed.bucket)
        sys.exit(2)

    act_on = bucket

    if parsed.file:
        k = bucket.get_key(parsed.file)

        if k is None:
            print(f"No {parsed.file} in {parsed.bucket}")
            sys.exit(2)

        act_on = k

    act_on.set_acl("public-read" if parsed.acl_public else "private")
    print("ACL updated")


def enable_static_site(get_client, args):
    """
    Turns a bucket into a static website
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " ws-create")

    parser.add_argument(
        "bucket",
        metavar="BUCKET",
        type=str,
        help="The bucket to turn into a static site",
    )
    parser.add_argument(
        "--ws-index",
        metavar="INDEX",
        type=str,
        help="The file to serve as the index of the website",
    )
    parser.add_argument(
        "--ws-error",
        metavar="ERROR",
        type=str,
        help="The file to serve as the error page of the website",
    )

    parsed = parser.parse_args(args)
    client = get_client()

    # get the bucket
    try:
        bucket = client.get_bucket(parsed.bucket)
    except S3ResponseError:
        print("No bucket named " + parsed.bucket)
        sys.exit(2)

    # make the site
    bucket.set_acl("public-read")
    bucket.configure_website(parsed.ws_index, parsed.ws_error)
    print(
        "Static site now available at "
        f"https://{parsed.bucket}.website-{client.obj_cluster}.linodeobjects.com"
    )


def static_site_info(get_client, args):
    """
    Returns info about a configured static site
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " ws-info")

    parser.add_argument(
        "bucket",
        metavar="BUCKET",
        type=str,
        help="The bucket to return static site information on.",
    )

    parsed = parser.parse_args(args)
    client = get_client()

    # get the bucket
    try:
        bucket = client.get_bucket(parsed.bucket)
    except S3ResponseError:
        print("No bucket named " + parsed.bucket)
        sys.exit(2)

    info = bucket.get_website_configuration()

    index = info["WebsiteConfiguration"]["IndexDocument"]["Suffix"]
    error = info["WebsiteConfiguration"]["ErrorDocument"]["Key"]

    web_template = BASE_WEBSITE_TEMPLATE.format(client.host.split(".")[0])
    print(f"Bucket {parsed.bucket}: Website configuration")
    print(f"Website endpoint: {parsed.bucket}.{web_template}")
    print(f"Index document: {index}")
    print(f"Error document: {error}")


def show_usage(get_client, args):
    """
    Shows space used by all buckets in this cluster, and total space
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " du")

    parser.add_argument(
        "bucket",
        metavar="BUCKET",
        type=str,
        nargs="?",
        help="Optional.  If given, only shows usage for that bucket. "
        "If omitted, shows usage for all buckets.",
    )

    parsed = parser.parse_args(args)
    client = get_client()

    if parsed.bucket:
        try:
            buckets = [client.get_bucket(parsed.bucket)]
        except S3ResponseError:
            print("No bucket named " + parsed.bucket)
            sys.exit(2)
    else:
        # all buckets
        buckets = client.get_all_buckets()

    grand_total = 0
    for b in buckets:
        # find total size of each
        total = num = 0
        for obj in b.list():
            num += 1
            total += obj.size

        grand_total += total

        total = _denominate(total)

        tab = _borderless_table([[_pad_to(total, length=7), f"{num} objects", b.name]])
        print(tab.table)

    if len(buckets) > 1:
        print("--------")
        print(f"{_denominate(grand_total)} Total")

    sys.exit(0)


def _denominate(total):
    """
    Coverts bucket size to human readable bytes.
    """
    total = float(total)
    denomination = ["KB", "MB", "GB", "TB"]
    for x in denomination:
        if total > 1024:
            total = total / 1024
        if total < 1024:
            total = round(total, 2)
            total = str(total) + " " + x
            break
    return total


def list_all_objects(get_client):
    """
    Lists all objects in all buckets
    """
    client = get_client()

    # all buckets
    buckets = client.get_all_buckets()

    for b in buckets:
        print()

        for obj in b.list():
            if isinstance(obj, Prefix):
                size = "DIR"
            else:
                size = obj.size

            print(
                f"{_convert_datetime(obj.last_modified) if size != 'DIR' else ' ' * 16} "
                f"{_pad_to(size, 9, right_align=True)}   "
                f"{b.name}/{obj.key}"
            )

    sys.exit(0)


def disable_static_site(get_client, args):
    """
    Disables static site for a bucket
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE + " du")

    parser.add_argument(
        "bucket",
        metavar="BUCKET",
        type=str,
        nargs="?",
        help="The bucket to disable static site for.",
    )

    parsed = parser.parse_args(args)
    client = get_client()

    # get the bucket
    try:
        bucket = client.get_bucket(parsed.bucket)
    except S3ResponseError:
        print("No bucket named " + parsed.bucket)
        sys.exit(2)

    # make the site
    bucket.delete_website_configuration()
    print(f"Website configuration deleted for {parsed.bucket}")


COMMAND_MAP = {
    "mb": create_bucket,
    "rb": delete_bucket,
    "ls": list_objects_or_buckets,
    "la": list_all_objects,
    "du": show_usage,
    "put": upload_object,
    "get": get_object,
    "rm": delete_object,
    "del": delete_object,
    #'sync': sync_dir, TODO - syncs a directory
    "signurl": generate_url,
    "setacl": set_acl,
    "ws-create": enable_static_site,
    "ws-info": static_site_info,
    "ws-delete": disable_static_site,
    #'info': get_info,
}


def call(args, context):  # pylint: disable=too-many-branches,too-many-statements
    """
    This is called when the plugin is invoked
    """
    if not HAS_BOTO:
        # we can't do anything - ask for an install
        pip_version = "pip3" if sys.version[0] == 3 else "pip"

        print(
            "This plugin requires the 'boto' module.  Please install it by running "
            f"'{pip_version} install boto'"
        )

        sys.exit(2)  # requirements not met - we can't go on

    parser = argparse.ArgumentParser(PLUGIN_BASE, add_help=False)
    parser.add_argument(
        "command",
        metavar="COMMAND",
        nargs="?",
        type=str,
        help="The command to execute in object storage",
    )
    parser.add_argument(
        "--cluster",
        metavar="CLUSTER",
        type=str,
        help="The cluster to use for the operation",
    )

    parsed, args = parser.parse_known_args(args)

    # don't mind --no-defaults if it's there; the top-level parser already took care of it
    try:
        args.remove("--no-defaults")
    except ValueError:
        pass

    if not parsed.command:
        # show help if invoked with no command
        parser.print_help()

        # additional help
        print()
        print("Available commands: ")

        command_help_map = [
            [name, func.__doc__.strip()] for name, func in sorted(COMMAND_MAP.items())
        ]

        tab = SingleTable(command_help_map)
        tab.inner_heading_row_border = False
        print(tab.table)
        print()
        print(
            "Additionally, you can regenerate your Object Storage keys using the "
            "'regenerate-keys' command or configure defaults for the plugin using "
            "the 'configure' command."
        )
        print()
        print("See --help for individual commands for more information")

        sys.exit(0)

    # make a client, but only if we weren't printing help

    access_key, secret_key = (
        os.environ.get(ENV_ACCESS_KEY_NAME, None),
        os.environ.get(ENV_SECRET_KEY_NAME, None),
    )

    if not "--help" in args:
        if access_key and not secret_key or secret_key and not access_key:
            print(
                f"You must set both {ENV_ACCESS_KEY_NAME} and {ENV_SECRET_KEY_NAME}, or neither"
            )
            sys.exit(1)

        # not given on command line, so look them up
        if not access_key:
            access_key, secret_key = _get_s3_creds(context.client)

    cluster = parsed.cluster
    if context.client.defaults:
        cluster = cluster or context.client.config.plugin_get_value("cluster")

    def get_client():
        """
        Get the boto client based on the cluster, or ask to configure a
        default cluster if one is not specified. This is in a method so
        command methods can do this work AFTER displaying help,
        that way help can be shown without specifying a cluster
        or having a valid OBJ key.
        """
        current_cluster = cluster
        if current_cluster is None and not context.client.defaults:
            print("Error: cluster is required.")
            sys.exit(1)

        if current_cluster is None:
            print(
                "Error: No default cluster is configured.  Either configure the CLI "
                "or invoke with --cluster to specify a cluster."
            )
            _configure_plugin(context.client)
            current_cluster = context.client.config.plugin_get_value("cluster")

        return _get_boto_client(current_cluster, access_key, secret_key)

    if parsed.command in COMMAND_MAP:
        try:
            COMMAND_MAP[parsed.command](get_client, args)
        except S3ResponseError as e:
            if e.error_code:
                print(f"Error: {e.error_code}")
            else:
                print(e)
            sys.exit(4)
        except S3CreateError as e:
            print(f"Error: {e}")
            sys.exit(5)
        except BotoClientError as e:
            message_parts = e.message.split(":")
            if len(message_parts) > 0:
                message = ":".join(message_parts[0:])
            else:
                message = e.message
            print(f"Error: {message}")
            sys.exit(6)
    elif parsed.command == "regenerate-keys":
        print("Regenerating Object Storage keys..")
        _get_s3_creds(context.client, force=True)
        print("Done.")
        print(
            "Warning: Your old Object Storage keys _were not_ automatically expired!  If you want "
            "to expire them, see `linode-cli object-storage keys-list` and "
            "`linode-cli object-storage keys-delete [KEYID]`."
        )
    elif parsed.command == "configure":
        _configure_plugin(context.client)
    else:
        print(f"No command {parsed.command}")
        sys.exit(1)


def _get_boto_client(cluster, access_key, secret_key):
    """
    Returns a boto client object that can be used to communicate with the Object
    Storage cluster.
    """
    client = boto.connect_s3(
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        host=BASE_URL_TEMPLATE.format(cluster),
        calling_format=OrdinaryCallingFormat(),
    )

    # set this for later use
    client.obj_cluster = cluster

    return client


def _get_s3_creds(client, force=False):
    """
    Retrieves stored s3 creds for the acting user from the config, or generates new
    creds using the client and stores them if none exist

    :param client: The client object from the invoking PluginContext
    :type client: linodecli.CLI
    :param force: If True, get new creds even if there are already creds stored.
                  This is used to rotate creds.
    :type force: bool

    :returns: The access key and secret key for this user
    :rtype: tuple(str, str)
    """
    access_key = client.config.plugin_get_value("access-key")
    secret_key = client.config.plugin_get_value("secret-key")

    if force or access_key is None:
        # this means there are no stored s3 creds for this user - set them up

        # but first - is there actually a config?  If we got this far, creds aren't
        # being provided by the environment, but if the CLI is running without a
        # config, we shouldn't generate new keys (or we'd end up doing so with each
        # request) - instead ask for them to be set up.
        if client.config.get_value("token") is None:
            print(
                "You are running the Linode CLI without a configuration file, but "
                "object storage keys were not configured.  "
                "Please set the following variables in your environment: "
                f"'{ENV_ACCESS_KEY_NAME}' and '{ENV_SECRET_KEY_NAME}'.  If you'd rather "
                "configure the CLI, unset the 'LINODE_CLI_TOKEN' environment "
                "variable and then run `linode-cli configure`."
            )
            sys.exit(1)

        # before we do anything, can they do object storage?
        status, resp = client.call_operation("account", "view")

        if status != 200:
            if status == 401:
                # special case - oauth token isn't allowed to do this
                print(NO_SCOPES_ERROR)
                sys.exit(4)
            if status == 403:
                # special case - restricted users can't use obj
                print(NO_ACCESS_ERROR)
                sys.exit(4)
            # something went wrong - give up
            print("Key generation failed!")
            sys.exit(4)

        # label caps at 50 characters - trim some stuff maybe
        # static characters in label account for 13 total
        # timestamp is 10 more
        # allow 13 characters both for username and hostname
        timestamp_part = str(time.time()).split(".", maxsplit=1)[0]
        truncated_user = getpass.getuser()[:13]
        truncated_hostname = socket.gethostname()[:13]

        creds_label = (
            f"linode-cli-{truncated_user}@{truncated_hostname}-{timestamp_part}"
        )

        if len(creds_label) > 50:
            # if this is somehow still too long, trim from the front
            creds_label = creds_label[50 - len(creds_label) :]

        status, resp = client.call_operation(
            "object-storage", "keys-create", ["--label", f"{creds_label}"]
        )

        if status != 200:
            if status == 401:
                # special case - oauth token isn't allowed to do this
                print(NO_SCOPES_ERROR)
                sys.exit(4)
            if status == 403:
                # special case - restricted users can't use obj
                print(NO_ACCESS_ERROR)
                sys.exit(4)
            # something went wrong - give up
            print("Key generation failed!")
            sys.exit(3)

        access_key = resp["access_key"]
        secret_key = resp["secret_key"]

        client.config.plugin_set_value("access-key", access_key)
        client.config.plugin_set_value("secret-key", secret_key)
        client.config.write_config(silent=True)

    return access_key, secret_key


def _configure_plugin(client):
    """
    Configures a default cluster value.
    """
    clusters = [
        c["id"]
        for c in client.config._do_get_request(  # pylint: disable=protected-access
            "/object-storage/clusters", token=client.config.get_value("token")
        )["data"]
    ]

    cluster = client.config._default_thing_input(  # pylint: disable=protected-access
        "Configure a default Cluster for operations.",
        clusters,
        "Default Cluster: ",
        "Please select a valid Cluster",
        optional=False,  # this is the only configuration right now
    )

    client.config.plugin_set_value("cluster", cluster)
    client.config.write_config()


def _progress(cur, total):
    """
    Draws the upload progress bar.
    """
    # We can't divide by zero :)
    if total == 0.0:
        return

    percent = f"{100 * (cur / float(total)):.1f}"
    progress = int(100 * cur // total)
    progress_bar = ("#" * progress) + ("-" * (100 - progress))
    print(f"\r |{progress_bar}| {percent}%", end="\r")

    if cur == total:
        print()


# helper functions for output
def _borderless_table(data):
    """
    Returns a terminaltables.SingleTable object with no borders and correct padding
    """
    tab = SingleTable(data)
    tab.inner_heading_row_border = False
    tab.inner_column_border = False
    tab.outer_border = False
    tab.padding_left = 0
    tab.padding_right = 2

    return tab


def _convert_datetime(datetime_str):
    """
    Given a string in INCOMING_DATE_FORMAT, returns a string in DATE_FORMAT
    """
    return datetime.strptime(datetime_str, INCOMING_DATE_FORMAT).strftime(DATE_FORMAT)


def _pad_to(val, length=10, right_align=False):  # pylint: disable=unused-argument
    """
    Pads val to be at minimum length characters long
    """
    ret = str(val)
    padding = ""

    if len(ret) < 10:
        padding = " " * (10 - len(ret))

    if right_align:
        ret = padding + ret
    else:
        ret = ret + padding

    return ret



"""
Handles configuring the cli, as well as loading configs so that they can be
used elsewhere.
"""

import re
import socket
import argparse
import webbrowser
from http import server

try:
    # python3
    import configparser
except ImportError:
    # python2
    import ConfigParser as configparser

import os
import sys

import requests

ENV_TOKEN_NAME = "LINODE_CLI_TOKEN"

LEGACY_CONFIG_DIR = os.path.expanduser("~")
LEGACY_CONFIG_NAME = ".linode-cli"
CONFIG_DIR = os.environ.get("XDG_CONFIG_HOME", f"{os.path.expanduser('~')}/.config")

CONFIG_NAME = "linode-cli"
TOKEN_GENERATION_URL = "https://cloud.linode.com/profile/tokens"

# This is used for web-based configuration
OAUTH_CLIENT_ID = "5823b4627e45411d18e9"

# this is a list of browser that _should_ work for web-based auth.  This is mostly
# intended to exclude lynx and other terminal browsers which could be opened, but
# won't work.
KNOWN_GOOD_BROWSERS = {
    "chrome",
    "firefox",
    "mozilla",
    "netscape",
    "opera",
    "safari",
    "chromium",
    "chromium-browser",
    "epiphany",
}

# in the event that we can't load the styled landing page from file, this will
# do as a landing page
DEFAULT_LANDING_PAGE = """
<h2>Success</h2><br/><p>You may return to your terminal to continue..</p>
<script>
// this is gross, sorry
let r = new XMLHttpRequest('http://localhost:{port}');
r.open('GET', '/token/'+window.location.hash.substr(1));
r.send();
</script>
"""


class CLIConfig:
    """
    Generates the necessary config for the Linode CLI
    """

    def __init__(self, base_url, username=None, skip_config=False):
        self.base_url = base_url
        self.username = username
        self.config = self._get_config(load=not skip_config)
        self.running_plugin = None
        self.used_env_token = False

        self._configured = False

        self.configure_with_pat = "--token" in sys.argv

        if (
            not skip_config
            and not self.config.has_option("DEFAULT", "default-user")
            and self.config.has_option("DEFAULT", "token")
        ):
            self._handle_no_default_user()

        environ_token = os.getenv(ENV_TOKEN_NAME, None)

        if (
            not self.config.has_option("DEFAULT", "default-user")
            and not skip_config
            and environ_token is None
        ):
            self.configure()
        elif environ_token is not None:
            self.used_env_token = True

    def set_user(self, username):
        """
        Sets the acting username.  If this username is not in the config, this is
        an error.  This overrides the default username
        """
        if not self.config.has_section(username):
            print(f"User {username} is not configured!")
            sys.exit(1)

        self.username = username

    def default_username(self):
        """
        Returns the default-user Username
        """
        if self.config.has_option("DEFAULT", "default-user"):
            return self.config.get("DEFAULT", "default-user")
        return ""

    def update_namespace(self, namespace, new_dict):
        """
        In order to update the namespace, we need to turn it into a dict, modify it there,
        then reconstruct it with the exploded dict.
        """
        ns_dict = vars(namespace)
        warn_dict = {}
        for k in new_dict:
            if k.startswith("plugin-"):
                # plugins set config options that start with 'plugin-' - these don't
                # get included in the updated namespace
                continue
            if k in ns_dict and isinstance(k, list):
                ns_dict[k].append(new_dict[k])
            if k in ns_dict and ns_dict[k] is None:
                warn_dict[k] = new_dict[k]
                ns_dict[k] = new_dict[k]
        if not any(x in ["--suppress-warnings", "--no-headers"] for x in sys.argv):
            print(
                f"using default values: {warn_dict}, use --no-defaults flag to disable defaults"
            )
        return argparse.Namespace(**ns_dict)

    def update(self, namespace, allowed_defaults):
        """
        This updates a Namespace (as returned by ArgumentParser) with config values
        if they aren't present in the Namespace already.
        """
        if self.used_env_token and self.config is None:
            # the CLI is using a token defined in the environment; as such, we may
            # not have actually loaded a config file.  That's fine, there are just
            # no defaults
            return None

        username = self.username or self.default_username()

        if not self.config.has_option(username, "token") and not os.environ.get(
            ENV_TOKEN_NAME, None
        ):
            print(f"User {username} is not configured.")
            sys.exit(1)

        if self.config.has_section(username) and allowed_defaults:
            # update_dicts = {
            #     default_key: self.config.get(username, default_key)
            #     for default_key in allowed_defaults
            #     if self.config.has_option(username, default_key)
            #     }
            update_dicts = {}
            for default_key in allowed_defaults:
                if not self.config.has_option(username, default_key):
                    continue
                value = self.config.get(username, default_key)
                if default_key == "authorized_users":
                    update_dicts[default_key] = [value]
                else:
                    update_dicts[default_key] = value
            return self.update_namespace(namespace, update_dicts)
        return namespace

    def get_token(self):
        """
        Returns the token for a configured user
        """
        if self.used_env_token:
            return os.environ.get(ENV_TOKEN_NAME, None)

        if self.config.has_option(self.username or self.default_username(), "token"):
            return self.config.get(self.username or self.default_username(), "token")
        return ""

    def remove_user(self, username):
        """
        Removes the requested user from the config.  If the user is the default,
        this exits with error
        """
        if self.default_username() == username:
            print(
                f"Cannot remote {username} as they are the default user! You can "
                "change the default user with: `linode-cli set-user USERNAME`"
            )
            sys.exit(1)

        if self.config.has_section(username):
            self.config.remove_section(username)
            self.write_config()

    def print_users(self):
        """
        Prints all users available and exits
        """
        print("Configured Users: ")
        default_user = self.default_username()

        for sec in self.config.sections():
            if sec != "DEFAULT":
                print(f'{"*" if sec == default_user else " "}  {sec}')

        sys.exit(0)

    def set_default_user(self, username):
        """
        Sets the default user.  If that user isn't in the config, exits with error
        """
        if not self.config.has_section(username):
            print(f"User {username} is not configured!")
            sys.exit(1)

        self.config.set("DEFAULT", "default-user", username)
        self.write_config()

    # plugin methods - these are intended for plugins to utilize to store their
    # own persistent config information
    def get_value(self, key):
        """
        Retrieves and returns an existing config value for the current user.  This
        is intended for plugins to use instead of having to deal with figuring out
        who the current user is when accessing their config.

        .. warning::
           Plugins _MUST NOT_ set values for the user's config except through
           ``plugin_set_value`` below.

        :param key: The key to look up.
        :type key: str

        :returns: The value for that key, or None if the key doesn't exist for the
                  current user.
        :rtype: any
        """
        username = self.username or self.default_username()

        if not self.config.has_option(username, key):
            return None

        return self.config.get(username, key)

    def plugin_set_value(self, key, value):
        """
        Sets a new config value for a plugin for the current user.  Plugin config
        keys are set in the following format::

           plugin-{plugin_name}-{key}

        Values set with this method are intended to be retrieved with ``plugin_get_value``
        below.

        :param key: The config key to set - this is needed to retrieve the value
        :type key: str
        :param value: The value to set for this key
        :type value: any
        """
        if self.running_plugin is None:
            raise RuntimeError("No running plugin to retrieve configuration for!")

        username = self.username or self.default_username()
        self.config.set(username, f"plugin-{self.running_plugin}-{key}", value)

    def plugin_get_value(self, key):
        """
        Retrieves and returns a config value previously set for a plugin.  Your
        plugin should have set this value in the past.  If this value does not
        exist in the config, ``None`` is returned.  This is the only time
        ``None`` is returned, so receiving this value should be treated as
        "plugin is not configured."

        :param key: The key of the value to return
        :type key: str

        :returns: The value for this plugin for this key, or None if not set
        :rtype: any
        """
        if self.running_plugin is None:
            raise RuntimeError("No running plugin to retrieve configuration for!")

        username = self.username or self.default_username()
        full_key = f"plugin-{self.running_plugin}-{key}"

        if not self.config.has_option(username, full_key):
            return None

        return self.config.get(username, full_key)

    def write_config(self, silent=False):
        """
        Saves the config file as it is right now.  This can be used by plugins
        to save values they've set, and is used internally to update the config
        on disk when a new user if configured.

        :param silent: If True, does not print a message noting the config file
                       has been updated.  This is primarily intended for silently
                       updated the config file from one version to another.
        :type silent: bool
        """

        # Create the ~/.config directory if it does not exist
        if not os.path.exists(f"{os.path.expanduser('~')}/.config"):
            os.makedirs(f"{os.path.expanduser('~')}/.config")

        with open(self._get_config_path(), "w", encoding="utf-8") as f:
            self.config.write(f)

        if not silent:
            print(f"\nConfig written to {self._get_config_path()}")

    def _username_for_token(self, token):
        """
        A helper function that returns the username assocaited with a token by
        requesting it from the API
        """
        u = self._do_get_request("/profile", token=token, exit_on_error=False)
        if "errors" in u:
            reasons = ",".join([c["reason"] for c in u["errors"]])
            print(f"That token didn't work: {reasons}")
            return None

        return u["username"]

    def _get_token_terminal(self):
        """
        Handles prompting the user for a Personal Access Token and checking it
        to ensure it works.
        """
        print(
            f"""
First, we need a Personal Access Token.  To get one, please visit
{TOKEN_GENERATION_URL} and click
"Create a Personal Access Token".  The CLI needs access to everything
on your account to work correctly."""
        )

        while True:
            token = input("Personal Access Token: ")

            username = self._username_for_token(token)
            if username is not None:
                break

        return username, token

    def _get_token_web(self):
        """
        Handles OAuth authentication for the CLI.  This requires us to get a temporary
        token over OAuth and then use it to create a permanent token for the CLI.
        This function returns the token the CLI should use, or exits if anything
        goes wrong.
        """
        temp_token = self._handle_oauth_callback()
        username = self._username_for_token(temp_token)

        if username is None:
            print("OAuth failed.  Please try again of use a token for auth.")
            sys.exit(1)

        # the token returned via public oauth will expire in 2 hours, which
        # isn't great.  Instead, we're gonna make a token that expires never
        # and store that.
        result = self._do_request(
            requests.post,
            "/profile/tokens",
            token=temp_token,
            # generate the actual token with a label like:
            #  Linode CLI @ linode
            # The creation date is already recoreded with the token, so
            # this should be all the relevant info.
            body={"label": f"Linode CLI @ {socket.gethostname()}"},
        )

        return username, result["token"]

    def _handle_oauth_callback(self):
        """
        Sends the user to a URL to perform an OAuth login for the CLI, then redirets
        them to a locally-hosted page that captures teh token
        """
        # load up landing page HTML
        landing_page_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "oauth-landing-page.html"
        )

        try:
            with open(landing_page_path, encoding="utf-8") as f:
                landing_page = f.read()
        except:
            landing_page = DEFAULT_LANDING_PAGE

        class Handler(server.BaseHTTPRequestHandler):
            """
            The issue here is that Login sends the token in the URL hash, meaning
            that we cannot see it on the server side.  An attempt was made to
            get the client (browser) to send an ajax request to pass it along,
            but that's pretty gross and also isn't working.  Needs more thought.
            """

            def do_GET(self):
                """
                get the access token
                """
                if "token" in self.path:
                    # we got a token!  Parse it out of the request
                    token_part = self.path.split("/", 2)[2]

                    m = re.search(r"access_token=([a-z0-9]+)", token_part)
                    if m and len(m.groups()):
                        self.server.token = m.groups()[0]

                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()

                # TODO: Clean up this page and make it look nice
                self.wfile.write(
                    bytes(
                        landing_page.format(port=self.server.server_address[1]).encode(
                            "utf-8"
                        )
                    )
                )

            def log_message(self, form, *args):  # pylint: disable=arguments-differ
                """Don't actually log the request"""

        # start a server to catch the response
        serv = server.HTTPServer(("localhost", 0), Handler)
        serv.token = None

        # figure out the URL to direct the user to and print out the prompt
        # pylint: disable-next=line-too-long
        url = f"https://login.linode.com/oauth/authorize?client_id={OAUTH_CLIENT_ID}&response_type=token&scopes=*&redirect_uri=http://localhost:{serv.server_address[1]}"
        print(
            f"""A browser should open directing you to this URL to authenticate:

{url}

If you are not automatically directed there, please copy/paste the link into your browser
to continue..
"""
        )

        webbrowser.open(url)

        try:
            while serv.token is None:
                # serve requests one at a time until we get a token or are interrupted
                serv.handle_request()
        except KeyboardInterrupt:
            print()
            print(
                "Giving up.  If you couldn't get web authentication to work, please "
                "try token using a token by invoking with `linode-cli configure --token`, "
                "and open an issue at https://github.com/linode/linode-cli"
            )
            sys.exit(1)

        return serv.token

    def configure(self):  # pylint: disable=too-many-branches,too-many-statements
        """
        This assumes we're running interactively, and prompts the user
        for a series of defaults in order to make future CLI calls
        easier.  This also sets up the config file.
        """
        # If configuration has already been done in this run, don't do it again.
        if self._configured:
            return
        config = {}
        # we're configuring the default user if there is no default user configured
        # yet
        is_default = not self.config.has_option("DEFAULT", "default-user")
        username = None
        token = None

        print(
            """Welcome to the Linode CLI.  This will walk you through some initial setup."""
        )

        if ENV_TOKEN_NAME in os.environ:
            print(
                """Using token from {env_token_name}.
Note that no token will be saved in your configuration file.
    * If you lose or remove {env_token_name}.
    * All profiles will use {env_token_name}.""".format(
                    env_token_name=ENV_TOKEN_NAME
                )
            )
            username = "DEFAULT"
            token = os.getenv(ENV_TOKEN_NAME)

        else:
            # let's see if we _can_ use web
            can_use_browser = True
            try:
                webbrowser.get()
            except webbrowser.Error:
                # there are no browsers installed
                can_use_browser = False

            if can_use_browser and not KNOWN_GOOD_BROWSERS.intersection(
                webbrowser._tryorder  # pylint: disable=protected-access
            ):
                print()
                print(
                    "This tool defaults to web-based authentication, however "
                    "no known-working browsers were found."
                )

                while True:
                    r = input("Try it anyway? [y/N]: ")
                    if r.lower() in "yn ":
                        can_use_browser = r.lower() == "y"
                        break

            if self.configure_with_pat or not can_use_browser:
                username, config["token"] = self._get_token_terminal()
            else:
                # pylint: disable=line-too-long
                print()
                print(
                    "The CLI will use its web-based authentication to log you in.  "
                    "If you prefer to supply a Personal Access Token, use `linode-cli configure --token`. "
                )
                print()
                input(
                    "Press enter to continue.  This will open a browser and proceed with authentication."
                )
                username, config["token"] = self._get_token_web()
                # pylint: enable=line-too-long

            token = config["token"]

        print()
        print(f"Configuring {username}")
        print()

        regions = [r["id"] for r in self._do_get_request("/regions")["data"]]
        types = [t["id"] for t in self._do_get_request("/linode/types")["data"]]
        images = [i["id"] for i in self._do_get_request("/images")["data"]]

        is_full_access = self._check_full_access(token)

        auth_users = []

        if is_full_access:
            auth_users = [
                u["username"]
                for u in self._do_get_request(
                    "/account/users", exit_on_error=False, token=token
                )["data"]
                if "ssh_keys" in u
            ]

        # get the preferred things
        config["region"] = self._default_thing_input(
            "Default Region for operations.",
            regions,
            "Default Region (Optional): ",
            "Please select a valid Region, or press Enter to skip",
        )

        config["type"] = self._default_thing_input(
            "Default Type of Linode to deploy.",
            types,
            "Default Type of Linode (Optional): ",
            "Please select a valid Type, or press Enter to skip",
        )

        config["image"] = self._default_thing_input(
            "Default Image to deploy to new Linodes.",
            images,
            "Default Image (Optional): ",
            "Please select a valid Image, or press Enter to skip",
        )

        if auth_users:
            config["authorized_users"] = self._default_thing_input(
                "Select the user that should be given default SSH access to new Linodes.",
                auth_users,
                "Default Option (Optional): ",
                "Please select a valid Option, or press Enter to skip",
            )

        # save off the new configuration
        if username != "DEFAULT" and not self.config.has_section(username):
            self.config.add_section(username)

        if not is_default:
            if username != self.default_username():
                while True:
                    value = input(
                        "Make this user the default when using the CLI? [y/N]: "
                    )

                    if value.lower() in "yn":
                        is_default = value.lower() == "y"
                        break
                    if not value.strip():
                        break

            if not is_default:  # they didn't change the default user
                print(
                    f"Active user will remain {self.config.get('DEFAULT', 'default-user')}"
                )

        if is_default:
            # if this is the default user, make it so
            self.config.set("DEFAULT", "default-user", username)
            print(f"Active user is now {username}")

        for k, v in config.items():
            if v:
                self.config.set(username, k, v)

        self.write_config()
        os.chmod(self._get_config_path(), 0o600)
        self._configured = True

    def _get_config_path(self):
        """
        Returns the path to the config file.
        """
        path = f"{LEGACY_CONFIG_DIR}/{LEGACY_CONFIG_NAME}"
        if os.path.exists(path):
            return path

        return f"{CONFIG_DIR}/{CONFIG_NAME}"

    def _get_config(self, load=True):
        """
        Returns a new ConfigParser object that represents the CLI's configuration.
        If load is false, we won't load the config from disk.

        :param load: If True, load the config from the default path.  Otherwise,
                     don't (and just return an empty ConfigParser)
        :type load: bool
        """
        conf = configparser.ConfigParser()

        if load:
            conf.read(self._get_config_path())

        return conf

    def _default_thing_input(
        self, ask, things, prompt, error, optional=True
    ):  # pylint: disable=too-many-arguments
        """
        Requests the user choose from a list of things with the given prompt and
        error if they choose something invalid.  If optional, the user may hit
        enter to not configure this option.
        """
        print(f"\n{ask}  Choices are:")
        for ind, thing in enumerate(things):
            print(f" {ind + 1} - {thing}")
        print()

        ret = ""
        while True:
            choice = input(prompt)

            if choice:
                try:
                    choice = int(choice)
                    choice = things[choice - 1]
                except:
                    pass

                if choice in list(things):
                    ret = choice
                    break
                print(error)
            else:
                if optional:
                    break
                print(error)
        return ret

    def _do_get_request(self, url, token=None, exit_on_error=True):
        """
        Does helper get requests during configuration
        """
        return self._do_request(
            requests.get, url, token=token, exit_on_error=exit_on_error
        )

    @staticmethod
    def _handle_response_status(response, exit_on_error=None):
        if 199 < response.status_code < 300:
            return

        print(f"Could not contact {response.url} - Error: {response.status_code}")
        if exit_on_error:
            sys.exit(4)

    def _do_request(
        self, method, url, token=None, exit_on_error=None, body=None
    ):  # pylint: disable=too-many-arguments
        """
        Does helper requests during configuration
        """
        headers = {}

        if token is not None:
            headers["Authorization"] = f"Bearer {token}"
            headers["Content-type"] = "application/json"

        result = method(self.base_url + url, headers=headers, json=body)

        self._handle_response_status(result, exit_on_error=exit_on_error)

        return result.json()

    def _check_full_access(self, token):
        headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }

        result = requests.get(
            self.base_url + "/profile/grants", headers=headers, timeout=120
        )

        self._handle_response_status(result, exit_on_error=True)

        return result.status_code == 204

    def _handle_no_default_user(self):
        """
        Handle the case that there is no default user in the config
        """
        users = [c for c in self.config.sections() if c != "DEFAULT"]

        if len(users) == 1:
            # only one user configured - they're the default
            self.config.set("DEFAULT", "default-user", users[0])
            self.write_config(silent=True)
            return

        if len(users) == 0:
            # config is new or _really_ old
            token = self.config.get("DEFAULT", "token")

            if token is not None:
                # there's a token in the config - configure that user
                u = self._do_get_request("/profile", token=token, exit_on_error=False)

                if "errors" in u:
                    # this token was bad - reconfigure
                    self.configure()
                    return

                # setup config for this user
                username = u["username"]

                self.config.set("DEFAULT", "default-user", username)
                self.config.add_section(username)
                self.config.set(username, "token", token)
                self.config.set(
                    username, "region", self.config.get("DEFAULT", "region")
                )
                self.config.set(username, "type", self.config.get("DEFAULT", "type"))
                self.config.set(username, "image", self.config.get("DEFAULT", "image"))
                self.config.set(
                    username,
                    "authorized_keys",
                    self.config.get("DEFAULT", "authorized_keys"),
                )

                self.write_config(silent=True)
            else:
                # got nothin', reconfigure
                self.configure()

            # this should be handled
            return

        # more than one user - prompt for the default
        print("Please choose the active user.  Configured users are:")
        for u in users:
            print(f" {u}")
        print()

        while True:
            username = input("Active user: ")

            if username in users:
                self.config.set("DEFAULT", "default-user", username)
                self.write_config()
                return
            print(f"No user {username}")



"""
Initialize plugins for the CLI
"""
import sys
from importlib import import_module
from os import listdir
from os.path import dirname

_available_files = listdir(dirname(__file__))
available_local = [
    f[:-3] for f in _available_files if f.endswith(".py") and f != "__init__.py"
]


def available(config):
    """
    Returns a list of plugins that are available
    """
    additional = []
    if config.config.has_option("DEFAULT", "registered-plugins"):
        registered_plugins = config.config.get("DEFAULT", "registered-plugins")

        additional = registered_plugins.split(",")

    return available_local + additional


def invoke(name, args, context):
    """
    Given the plugin name, executes a plugin
    """
    # setup config to know what plugin is running
    context.client.config.running_plugin = name

    if name in available_local:
        plugin = import_module("linodecli.plugins." + name)
        plugin.call(args, context)
    elif name in available(context.client.config):
        # this is a third-party plugin
        try:
            plugin_module_name = context.client.config.config.get(
                "DEFAULT", f"plugin-name-{name}"
            )
        except KeyError:
            print(f"Plugin {name} is misconfigured - please re-register it")
            sys.exit(9)
        try:
            plugin = import_module(plugin_module_name)
        except ImportError:
            print(
                f"Expected module '{plugin_module_name}' not found.  "
                "Either {name} is misconfigured, or the backing module was uninstalled."
            )
            sys.exit(10)
        plugin.call(args, context)
    else:
        raise ValueError("No plugin named {name}")


class PluginContext:  # pylint: disable=too-few-public-methods
    """
    This class contains all context information provided to plugins when invoked.
    This includes access to the underlying CLI object to access the user's account,
    and the CLI access token the user has provided.
    """

    def __init__(self, token, client):
        """
        Constructs a new PluginContext with the given information
        """
        self.token = token
        self.client = client



"""
Classes related to OpenAPI-defined operations and their arguments and parameters.
"""

import argparse
import glob
import json
import platform
from getpass import getpass
from os import environ, path


def parse_boolean(value):
    """
    A helper to allow accepting booleans in from argparse.  This is intended to
    be passed to the `type=` kwarg for ArgumentParser.add_argument.
    """
    if value.lower() in ("yes", "true", "y", "1"):
        return True
    if value.lower() in ("no", "false", "n", "0"):
        return False
    raise argparse.ArgumentTypeError("Expected a boolean value")


def parse_dict(value):
    """
    A helper function to decode incoming JSON data as python dicts.  This is
    intended to be passed to the `type=` kwarg for ArgumentParaser.add_argument.
    """
    if not isinstance(value, str):
        raise argparse.ArgumentTypeError("Expected a JSON string")
    try:
        return json.loads(value)
    except Exception as e:
        raise argparse.ArgumentTypeError("Expected a JSON string") from e


class PasswordPromptAction(argparse.Action):
    """
    A special argparse Action to handle prompting for password.  Also accepts
    passwords on the terminal to allow for backwards-compatible behavior.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        # if not provided on the command line, pull from the environment if it
        # exists at this key
        environ_key = f"LINODE_CLI_{self.dest.upper()}"

        if values:
            if isinstance(values, str):
                password = values
            else:
                raise argparse.ArgumentTypeError(
                    "Expected a string (or leave blank for prompt)"
                )
        elif environ_key in environ:
            password = environ.get(environ_key)
        else:
            prompt = f"Value for {self.dest}: "
            password = getpass(prompt)
        setattr(namespace, self.dest, password)


class OptionalFromFileAction(argparse.Action):
    """
    A special action for handling loading a value from a file.  This will
    attempt to load the value from a file if the value looks like a path and
    the file exists, otherwise it will fall back to using the provided value.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        if isinstance(values, str):
            input_path = path.expanduser(values)

            # Windows doesn't natively expand globs, so we should implement it here
            if platform.system() == "Windows" and "*" in input_path:
                results = glob.glob(input_path, recursive=True)

                if len(results) < 1:
                    raise argparse.ArgumentError(
                        self, f"File matching pattern {input_path} not found"
                    )

                input_path = results[0]

            if path.exists(input_path) and path.isfile(input_path):
                with open(input_path, encoding="utf-8") as f:
                    data = f.read()
                setattr(namespace, self.dest, data)
            else:
                setattr(namespace, self.dest, values)
        else:
            raise argparse.ArgumentTypeError("Expected a string")


TYPES = {
    "string": str,
    "integer": int,
    "boolean": parse_boolean,
    "array": list,
    "object": parse_dict,
    "number": float,
}


class CLIArg:  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """
    An argument passed to the CLI with a flag, such as `--example value`.  These
    are defined in a requestBody in the api spec.
    """

    def __init__(
        self, name, arg_type, description, path, arg_format, list_item=None
    ):  # pylint: disable=too-many-arguments,redefined-outer-name
        self.name = name
        self.arg_type = arg_type
        self.arg_format = arg_format
        self.description = description.replace("\n", "").replace("\r", "")
        self.path = path
        self.arg_item_type = None  # populated during baking for arrays
        self.required = False  # this is set during baking
        self.list_item = list_item


class URLParam:  # pylint: disable=too-few-public-methods
    """
    An argument passed to the CLI positionally. These are defined in a path in
    the OpenAPI spec, in a "parameters" block
    """

    def __init__(self, name, param_type):
        self.name = name
        self.param_type = param_type

    def clone(self):
        """
        Returns a new URLParam that is exactly like this one
        """
        return URLParam(self.name, self.param_type)


class CLIOperation:  # pylint: disable=too-many-instance-attributes
    """
    A single operation described by the OpenAPI spec.  An operation is a method
    on a path, and should have a unique operationId to identify it.  Operations
    are responsible for parsing their own arguments and processing their
    responses with the help of their ResponseModel
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        command,
        action,
        method,
        url,
        summary,
        args,
        response_model,
        params,
        servers,
        docs_url=None,
        allowed_defaults=None,
        action_aliases=None,
    ):
        self.command = command
        self.action = action
        self.method = method
        self._url = url
        self.summary = summary
        self.args = args
        self.response_model = response_model
        self.params = params
        self.servers = servers
        self.docs_url = docs_url
        self.allowed_defaults = allowed_defaults
        self.action_aliases = action_aliases or []

    @property
    def url(self):
        """
        Returns the full URL for this resource based on servers and endpoint
        """
        base_url = self.servers[0]
        return base_url + "/" + self._url

    def parse_args(
        self, args
    ):  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        """
        Given sys.argv after the operation name, parse args based on the params
        and args of this operation
        """
        list_items = []

        #  build an argparse
        parser = argparse.ArgumentParser(
            prog=f"linode-cli {self.command} {self.action}",
            description=self.summary,
        )
        for param in self.params:
            parser.add_argument(
                param.name, metavar=param.name, type=TYPES[param.param_type]
            )

        if self.method == "get":
            # build args for filtering
            for attr in self.response_model.attrs:
                if attr.filterable:
                    expected_type = TYPES[attr.datatype]
                    if expected_type == list:
                        parser.add_argument(
                            "--" + attr.name,
                            type=TYPES[attr.item_type],
                            metavar=attr.name,
                            nargs="?",
                        )
                    else:
                        parser.add_argument(
                            "--" + attr.name, type=expected_type, metavar=attr.name
                        )

        elif self.method in ("post", "put"):
            # build args for body JSON
            for arg in self.args:
                if arg.arg_type == "array":
                    # special handling for input arrays
                    parser.add_argument(
                        "--" + arg.path,
                        metavar=arg.name,
                        action="append",
                        type=TYPES[arg.arg_item_type],
                    )
                elif arg.list_item is not None:
                    parser.add_argument(
                        "--" + arg.path,
                        metavar=arg.name,
                        action="append",
                        type=TYPES[arg.arg_type],
                    )
                    list_items.append((arg.path, arg.list_item))
                else:
                    if arg.arg_type == "string" and arg.arg_format == "password":
                        # special case - password input
                        parser.add_argument(
                            "--" + arg.path, nargs="?", action=PasswordPromptAction
                        )
                    elif arg.arg_type == "string" and arg.arg_format in (
                        "file",
                        "ssl-cert",
                        "ssl-key",
                    ):
                        parser.add_argument(
                            "--" + arg.path,
                            metavar=arg.name,
                            action=OptionalFromFileAction,
                            type=TYPES[arg.arg_type],
                        )
                    else:
                        parser.add_argument(
                            "--" + arg.path, metavar=arg.name, type=TYPES[arg.arg_type]
                        )

        parsed = parser.parse_args(args)
        lists = {}
        # group list items as expected
        for arg_name, list_name in list_items:
            item_name = arg_name.split(list_name)[1][1:]
            if hasattr(parsed, arg_name):
                val = getattr(parsed, arg_name) or []
                if not val:
                    continue
                if list_name not in lists:
                    new_list = [{item_name: c} for c in val]
                    lists[list_name] = new_list
                else:
                    update_list = lists[list_name]
                    for obj, item in zip(update_list, val):
                        obj[item_name] = item

        # break out list items with periods in their name into objects.  This
        # allows supporting nested lists
        for _, cur_list in lists.items():
            # for each list in lists
            for item in cur_list:
                # for each item in the list (these are dicts)
                new_dicts = {}
                remove_keys = []
                for k, v in item.items():
                    # if there's a period in the key, split it into a dict and
                    # possibly merge it with a dict that came from a prior split
                    #
                    # XXX: This only supports one layer of nested dicts in lists
                    if "." in k:
                        dict_key, key = k.split(".", 1)
                        if dict_key in new_dicts:
                            new_dicts[dict_key][key] = v
                        else:
                            new_dicts[dict_key] = {key: v}
                        remove_keys.append(k)

                # remove the original keys
                for key in remove_keys:
                    del item[key]
                # and add the combined keys
                item.update(new_dicts)

        # don't send along empty lists
        to_delete = []
        for k, v in lists.items():
            if len(v) == 0:
                to_delete.append(k)

        for c in to_delete:
            del lists[c]

        if lists:
            parsed = vars(parsed)
            parsed.update(lists)
            for name, _ in list_items:
                del parsed[name]
            parsed = argparse.Namespace(**parsed)

        return parsed

    def process_response_json(
        self, json, handler
    ):  # pylint: disable=redefined-outer-name
        """
        Processes the response as JSON and prints
        """
        if self.response_model is None:
            return

        json = self.response_model.fix_json(json)

        handler.print(self.response_model, json)



"""
This plugin allows easy uploading of new Images to Linode's Images system.

Usage:

   linode-cli image-upload --region us-east --label my-image /path/to/image.gz
"""

import argparse
import glob
import os
import platform
import sys

import requests

PLUGIN_BASE = "linode-cli image-upload"
MAX_UPLOAD_SIZE = 5 * 1024 * 1024 * 1024  # 5GB


def _progress(cur, total):
    """
    Draws the upload progress bar.
    """
    percent = f"{100 * (cur / float(total)):.1f}"
    progress = int(100 * cur // total)
    progress_bar = ("#" * progress) + ("-" * (100 - progress))
    print(f"\r |{progress_bar}| {percent}%", end="", flush=True)


class UploadProgressHelper:
    """
    Handles streaming uploads with python's requests library while printing a
    progress bar.
    """

    def __init__(self, filepath, chunk_size=5242880):
        self.filepath = filepath
        self.chunk_size = chunk_size
        self.read_size = 0
        self.total_size = os.path.getsize(filepath)

    def __iter__(self):
        """
        This is called by requests to get one chunk of data to upload.  Uploads
        are streamed from disk in chunks of the given size.  This will also update
        the progress bar.
        """
        with open(self.filepath, "rb") as upload_file:
            while True:
                data = upload_file.read(self.chunk_size)
                if not data:
                    _progress(self.total_size, self.total_size)
                    break
                self.read_size += len(data)
                _progress(self.read_size, self.total_size)
                yield data

    def __len__(self):
        return self.total_size


def call(args, context):
    """
    The entrypoint for this plugin
    """
    parser = argparse.ArgumentParser(PLUGIN_BASE, add_help=True)

    parser.add_argument(
        "--region",
        metavar="REGION",
        nargs="?",
        help="The region to upload the image to.  The uploaded "
        "image will be available to deploy in all regions, "
        "but initial deploys will be faster in the same region. "
        "Uploads will be faster to regions closer to you. "
        "Your default configured region is used if this is "
        "omitted.",
    )
    parser.add_argument(
        "--label",
        metavar="LABEL",
        nargs="?",
        help="Label for the new Image.  If omitted, the filename " "will be used.",
    )
    parser.add_argument(
        "--description",
        metavar="DESC",
        nargs="?",
        help="A description for this Image.  Blank if omitted.",
    )
    parser.add_argument(
        "file",
        metavar="FILE",
        help="The image file to upload.  Should be a raw disk image "
        "compressed with gzip, in .img.gz format.  We recommend "
        "using unpartitioned images with an ext3 or ext4 filesystem.",
    )

    parsed = parser.parse_args(args)

    # get default region populated
    context.client.config.update(parsed, ["region"])

    # make sure the file exists and is ready to upload
    filepath = os.path.expanduser(parsed.file)

    # Windows doesn't natively expand globs, so we should implement it here
    if platform.system() == "Windows" and "*" in filepath:
        results = glob.glob(filepath, recursive=True)

        if len(results) < 1:
            print(f"No file found matching pattern {filepath}")
            sys.exit(2)

        if len(results) > 1:
            print(
                f"warn: Found multiple files matching pattern {filepath}, using {results[0]}"
            )

        filepath = results[0]

    if not os.path.isfile(filepath):
        print(f"No file at {filepath}; must be a path to a valid file.")
        sys.exit(2)

    # make sure it's not larger than the max upload size
    if os.path.getsize(filepath) > MAX_UPLOAD_SIZE:
        print(f"File {filepath} is too large; compressed size must be less than 5GB")
        sys.exit(2)

    if not parsed.region:
        print("No region provided.  Please set a default region or use --region")
        sys.exit(1)

    label = parsed.label or os.path.basename(filepath)

    # generate an upload URL
    call_args = ["--region", parsed.region, "--label", label]
    if parsed.description:
        call_args += ["--description", parsed.description]

    status, resp = context.client.call_operation("images", "upload", call_args)

    if status != 200:
        if status == 401:
            print(
                "Your token was not authorized to use this endpoint.  Please "
                "reconfigure the CLI with `linode-cli configure` to ensure you "
                "can make this request."
            )
            sys.exit(3)
        if status == 404:
            print(
                "It looks like you are not in the Machine Images Beta, and therefore "
                "cannot upload images yet.  Please stay tuned, or open a support ticket "
                "to request access."
            )
            sys.exit(4)
        print(f"Upload failed with status {status}; response was {resp}")
        sys.exit(3)

    # grab the upload URL and image data
    image = resp["image"]
    upload_url = resp["upload_to"]

    # attempt to do the upload - use a streaming upload so that large files won't
    # be loaded entirely into memory
    requests.put(
        upload_url,
        headers={
            "Content-type": "application/octet-stream",
        },
        data=UploadProgressHelper(filepath),
        timeout=120,
    )
    print()

    # supposing it all worked, show the new Image
    context.client.handle_command("images", "view", [str(image["id"])])



"""
Classes related to parsing responses and display output from the API, based
on the OpenAPI spec.
"""

import os
import platform

DO_COLORS = True
# !! Windows compatibility for ANSI color codes !!
#
# If we're running on windows, we need to run the "color" command to enable
# ANSI color code support.
if platform.system() == "Windows":
    ver = platform.version()

    if "." in ver:
        ver = ver.split(".", 1)[0]

    try:
        verNum = int(ver)
    except ValueError:
        DO_COLORS = False

    # windows 10+ supports ANSI color codes after running the 'color' command to
    # properly set up the command prompt.  Older versions of windows do not, and
    # we should not attempt to use them there.
    if verNum >= 10:
        os.system("color")
    else:
        DO_COLORS = False


CLEAR_COLOR = "\x1b[0m"
COLOR_CODE_MAP = {
    "red": "\x1b[31m",
    "green": "\x1b[32m",
    "yellow": "\x1b[33m",
    "black": "\x1b[30m",
    "white": "\x1b[40m",
}


def colorize_string(string, color):
    """
    Returns the requested string, wrapped in ANSI color codes to colorize it as
    requested.  On platforms where colors are not supported, this just returns
    the string passed into it.
    """
    if not DO_COLORS:
        return string

    col = COLOR_CODE_MAP.get(color, CLEAR_COLOR)

    return f"{col}{string}{CLEAR_COLOR}"


class ModelAttr:  # pylint: disable=too-many-instance-attributes
    """
    Collects values from models
    """

    def __init__(  # pylint: disable=too-many-arguments
        self, name, filterable, display, datatype, color_map=None, item_type=None
    ):
        self.name = name
        self.value = None
        self.filterable = filterable
        self.display = display
        self.column_name = self.name.split(".")[-1]
        self.datatype = datatype
        self.color_map = color_map
        self.item_type = item_type

    def _get_value(self, model):
        """
        Returns the raw value from a model
        """
        # walk down json paths to find the value
        value = model
        for part in self.name.split("."):
            if value is None or value == {}:
                return None
            value = value[part]

        return value

    def render_value(self, model, colorize=True):
        """
        Given the model returned from the API, returns the correctly- rendered
        version of it.  This can transform text based on various rules
        configured in the spec using custom tags.  Currently supported tags:

        x-linode-cli-color
          A list of key-value pairs that represent the value, and its ideal color.
          The key "default_" is used to colorize anything that is not included.
          If omitted, no color is applied.
        """
        value = self._get_value(model)

        if isinstance(value, list):
            value = ", ".join([str(c) for c in value])

        if colorize and self.color_map is not None:
            # apply colors
            value = str(value)  # just in case
            color = self.color_map.get(value) or self.color_map["default_"]
            value = colorize_string(value, color)

        if value is None:
            # don't print the word "None"
            value = ""

        return value

    def get_string(self, model):
        """
        Returns the raw value from a model, cleaning up Nones and other values
        that won't render properly as strings
        """
        value = self._get_value(model)

        if value is None:
            value = ""
        elif isinstance(value, list):
            value = " ".join([str(c) for c in value])
        else:
            value = str(value)

        return value


class ResponseModel:  # pylint: disable=too-few-public-methods
    """
    Formats data for the response
    """

    def __init__(self, attrs, rows=None, nested_list=None):
        self.attrs = attrs
        self.rows = rows
        self.nested_list = nested_list

    def fix_json(self, json):  # pylint: disable=too-many-branches
        """
        Takes JSON from the API and formats it into a list of rows
        """
        if self.rows:  # pylint: disable=no-else-return
            # take the columns as specified
            ret = []
            for c in self.rows:
                cur = json
                for part in c.split("."):
                    cur = cur.get(part)

                if not cur:
                    # probably shouldn't happen, but ok
                    continue

                if isinstance(cur, list):
                    ret += cur
                else:
                    ret.append(cur)

            # we're good
            return ret
        elif self.nested_list:
            # we need to explode the rows into one row per entry in the nested list,
            # copying the external values
            if "pages" in json:
                json = json["data"]

            nested_lists = [c.strip() for c in self.nested_list.split(",")]
            ret = []

            for nested_list in nested_lists:
                path_parts = nested_list.split(".")

                if not isinstance(json, list):
                    json = [json]
                for cur in json:

                    nlist_path = cur
                    for p in path_parts:
                        nlist_path = nlist_path.get(p)
                    nlist = nlist_path

                    for item in nlist:
                        cobj = {k: v for k, v in cur.items() if k != path_parts[0]}
                        cobj["_split"] = path_parts[-1]
                        cobj[path_parts[0]] = item
                        ret.append(cobj)

            return ret
        elif "pages" in json:
            return json["data"]
        else:
            return [json]



"""
Responsible for managing spec and routing commands to operations.
"""

import pickle
import json
import sys
import re
import os
from distutils.version import LooseVersion, StrictVersion # pylint: disable=deprecated-module
from string import Template
from sys import stderr, version_info

import requests

from .configuration import CLIConfig
from .operation import CLIArg, CLIOperation, URLParam
from .output import OutputHandler, OutputMode
from .response import ModelAttr, ResponseModel

METHODS = ("get", "post", "put", "delete")
PIP_CMD = "pip3" if version_info.major == 3 else "pip"


class CLI:  # pylint: disable=too-many-instance-attributes
    """
    Responsible for loading or baking a spec and handling incoming commands
    """

    def __init__(self, version, base_url, skip_config=False):
        self.ops = {}
        self.spec = {}
        self.defaults = True  # whether to use default values for arguments
        self.page = 1
        self.page_size = 100
        self.debug_request = False
        self.version = version
        self.base_url = base_url
        self.spec_version = "None"
        self.suppress_warnings = False

        self.output_handler = OutputHandler()
        self.config = CLIConfig(self.base_url, skip_config=skip_config)
        self.load_baked()

    def _resolve_allOf(self, node):
        """
        Given the contents of an "allOf" node, returns the entire dct having parsed
        all refs and combined all other nodes.

        :param node: The contents of an 'allOf'
        :type node: list
        """
        ret = {}

        for cur in node:
            data = cur
            if "$ref" in cur:
                data = self._resolve_ref(cur["$ref"])
            props = {}
            if "properties" in data:
                props = data["properties"]
            elif "$ref" in cur and "/properties/" in cur["$ref"]:
                # if we referenced a property, we got a property
                props = data
            else:
                print(f"Warning: Resolved empty node for {cur} in {node}")
            ret.update(props)
        return ret

    def _resolve_ref(self, ref):
        """
        Resolves a reference to the referenced component.

        :param ref: A reference path, like '#/components/schemas/Linode'
        :type ref: str

        :returns: The resolved reference
        :rtype: dct
        """
        path_parts = ref.split("/")[1:]
        tmp = self.spec
        for part in path_parts:
            tmp = tmp[part]

        return tmp

    def _parse_args(
        self, node, prefix=None, args=None
    ):  # pylint: disable=too-many-branches
        """
        Given a node in a requestBody, parses out the properties and returns the
        CLIArg info
        """
        if args is None:
            args = {}
        if prefix is None:
            prefix = []

        for arg, info in node.items():
            if "allOf" in info:
                info = self._resolve_allOf(info["allOf"])
            while "$ref" in info:
                info = self._resolve_ref(info["$ref"])
            if "properties" in info:
                self._parse_args(info["properties"], prefix=prefix + [arg], args=args)
                continue  # we can't edit this level of the tree
            if info.get("readOnly"):
                continue
            if "$ref" in info:
                info = self._resolve_ref(info["$ref"])
            path = ".".join(prefix + [arg])
            args[path] = {
                "type": info.get("type") or "string",
                "desc": info.get("description") or "",
                "name": arg,
                "format": info.get("x-linode-cli-format", info.get("format", None)),
            }

            # if this is coming in as json, stop here
            if args[path]["format"] == "json":
                args[path]["type"] = "object"
                continue

            # handle input lists
            if args[path]["type"] == "array" and "items" in info:
                items = info["items"]

                if "allOf" in items:
                    # if items contain an "allOf", parse it down and format it
                    # as is expected here
                    items = self._resolve_allOf(items["allOf"])
                    items = {"type": "object", "items": items}
                if "$ref" in items:
                    # if it's just a ref, parse that out too
                    items = self._resolve_ref(items["$ref"])

                args[path]["item_type"] = items["type"]

                if (
                    items["type"] == "object"
                    and "properties" in items
                    and not items.get("readOnly")
                ):
                    # this is a special case - each item has its own properties
                    # that we need to capture separately
                    item_args = self._parse_args(
                        items["properties"], prefix=prefix + [arg]
                    )
                    for _, v in item_args.items():
                        v["list_item"] = path
                    args.update(item_args)
                    del args[path]  # remove the base element, which is junk

        return args

    def _parse_properties(self, node, prefix=None):
        """
        Given the value of a "properties" node, parses out the attributes and
        returns them as a list
        """
        if prefix is None:
            prefix = []
        attrs = []
        for name, info in node.items():
            if "properties" in info:
                attrs += self._parse_properties(info["properties"], prefix + [name])
            else:
                item_type = None
                item_container = info.get("items")
                if item_container:
                    item_type = item_container.get("type")
                attrs.append(
                    ModelAttr(
                        ".".join(prefix + [name]),
                        info.get("x-linode-filterable") or False,
                        info.get("x-linode-cli-display") or False,
                        info.get("type") or "string",
                        color_map=info.get("x-linode-cli-color"),
                        item_type=item_type,
                    )
                )

        return attrs

    def bake(
        self, spec
    ):  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        """
        Generates ops and bakes them to a pickle
        """
        self.spec = spec
        self.ops = {}
        default_servers = [c["url"] for c in spec["servers"]]

        for path, data in self.spec[ # pylint: disable=too-many-nested-blocks
            "paths"
        ].items():  # pylint: disable=too-many-nested-blocks
            command = data.get("x-linode-cli-command") or "default"
            if command not in self.ops:
                self.ops[command] = {}

            params = []
            if "parameters" in data:
                for info in data["parameters"]:
                    if "$ref" in info:
                        info = self._resolve_ref(info["$ref"])
                    params.append(URLParam(info["name"], info["schema"]["type"]))
            for m in METHODS:
                if m in data:
                    if data[m].get("x-linode-cli-skip"):
                        # some actions aren't available to the CLI - skip them
                        continue

                    action = data[m].get("x-linode-cli-action") or data[m].get(
                        "operationId"
                    )

                    if action is None:
                        print(f"warn: no operationId for {m.upper()} {path}")
                        continue

                    action_aliases = None

                    if isinstance(action, list):
                        if len(action) < 1:
                            print(f"warn: empty list for action {m.upper()}")
                            continue

                        action_aliases = action[1:]
                        action = action[0]

                    summary = data[m].get("summary") or ""

                    # Resolve the documentation URL
                    docs_url = None
                    tags = data[m].get("tags")
                    if tags is not None and len(tags) > 0 and len(summary) > 0:
                        tag_path = self._flatten_url_path(tags[0])
                        summary_path = self._flatten_url_path(summary)
                        docs_url = f"https://www.linode.com/docs/api/{tag_path}/#{summary_path}"

                    use_servers = (
                        [c["url"] for c in data[m]["servers"]]
                        if "servers" in data[m]
                        else default_servers
                    )

                    args = {}
                    required_fields = []
                    allowed_defaults = None
                    if m in ("post", "put") and "requestBody" in data[m]:
                        allowed_defaults = data[m]["requestBody"].get(
                            "x-linode-cli-allowed-defaults", None
                        )

                        if "application/json" in data[m]["requestBody"]["content"]:
                            body_schema = data[m]["requestBody"]["content"][
                                "application/json"
                            ]["schema"]

                            if "required" in body_schema:
                                required_fields = body_schema["required"]

                            if "allOf" in body_schema:
                                body_schema = self._resolve_allOf(body_schema["allOf"])
                            if "required" in body_schema:
                                required_fields += body_schema["required"]
                            if "$ref" in body_schema:
                                body_schema = self._resolve_ref(body_schema["$ref"])
                            if "required" in body_schema:
                                required_fields += body_schema["required"]
                            if "properties" in body_schema:
                                body_schema = body_schema["properties"]
                            if "required" in body_schema:
                                required_fields += body_schema["required"]

                            args = self._parse_args(body_schema, args={})

                    response_model = None
                    if (
                        "200" in data[m]["responses"]
                        and "application/json" in data[m]["responses"]["200"]["content"]
                    ):
                        resp_con = data[m]["responses"]["200"]["content"][
                            "application/json"
                        ]["schema"]

                        if (
                            "x-linode-cli-use-schema"
                            in data[m]["responses"]["200"]["content"][
                                "application/json"
                            ]
                        ):
                            # this body is atypical, and defines its own columns
                            # using this schema instead of the normal one.  This
                            # is usually pairs with x-linode-cli-rows so to handle
                            # endpoints that returns irregularly formatted data
                            resp_con = data[m]["responses"]["200"]["content"][
                                "application/json"
                            ]["x-linode-cli-use-schema"]

                        if "$ref" in resp_con:
                            resp_con = self._resolve_ref(resp_con["$ref"])
                        if "allOf" in resp_con:
                            resp_con.update(self._resolve_allOf(resp_con["allOf"]))
                        # handle pagination envelope
                        if (
                            "properties" in resp_con
                            and "pages" in resp_con["properties"]
                        ):
                            resp_con = resp_con["properties"]
                        if "pages" in resp_con and "data" in resp_con:
                            if "$ref" in resp_con["data"]["items"]:
                                resp_con = self._resolve_ref(
                                    resp_con["data"]["items"]["$ref"]
                                )
                            else:
                                resp_con = resp_con["data"]["items"]

                        attrs = []
                        if "properties" in resp_con:
                            attrs = self._parse_properties(resp_con["properties"])
                            # maybe we have special columns?
                            rows = (
                                data[m]["responses"]["200"]["content"][
                                    "application/json"
                                ].get("x-linode-cli-rows")
                                or None
                            )
                            nested_list = (
                                data[m]["responses"]["200"]["content"][
                                    "application/json"
                                ].get("x-linode-cli-nested-list")
                                or None
                            )
                            response_model = ResponseModel(
                                attrs, rows=rows, nested_list=nested_list
                            )

                    cli_args = []

                    for arg, info in args.items():
                        new_arg = CLIArg(
                            info["name"],
                            info["type"],
                            info["desc"].split(".")[0] + ".",
                            arg,
                            info["format"],
                            list_item=info.get("list_item"),
                        )

                        if arg in required_fields:
                            new_arg.required = True

                        # handle arrays
                        if "item_type" in info:
                            new_arg.arg_item_type = info["item_type"]
                        cli_args.append(new_arg)

                    # looks for param names that will be obscured by args
                    # clone the params since they're shared by all methods in this
                    # path, and we only want to modify this method's params
                    use_params = [c.clone() for c in params]
                    use_path = path
                    for p in use_params:
                        if p.name in args:
                            # or (m == 'get' and p.name in model_attrs):
                            # if we found a parameter name that is also and argument name
                            # append an underscore to both the parameter name and the
                            # parameter name in the URL
                            use_path = use_path.replace(
                                "{" + p.name + "}", "{" + p.name + "_}"
                            )
                            p.name += "_"

                    self.ops[command][action] = CLIOperation(
                        command,
                        action,
                        m,
                        use_path,
                        summary,
                        cli_args,
                        response_model,
                        use_params,
                        use_servers,
                        docs_url=docs_url,
                        allowed_defaults=allowed_defaults,
                        action_aliases=action_aliases,
                    )

        # remove any empty commands (those that have no actions)
        to_remove = []
        for command, actions in self.ops.items():
            if len(actions) == 0:
                to_remove.append(command)

        for command in to_remove:
            del self.ops[command]

        # hide the base_url from the spec away
        self.ops["_base_url"] = spec["servers"][0]["url"]
        self.ops["_spec_version"] = spec["info"]["version"]

        # finish the baking
        data_file = self._get_data_file()
        with open(data_file, "wb") as f:
            pickle.dump(self.ops, f)

    def get_fish_completions(self):
        """
        Generates and returns fish shell completions based on the baked spec
        """
        completion_template = Template(
            """# This is a generated file! Do not modify!
complete -c linode-cli -n "not __fish_seen_subcommand_from $subcommands" -x -a '$subcommands --help'
$command_items"""
        )

        command_template = Template(
            """complete -c linode-cli -n "__fish_seen_subcommand_from $command" \
                    -x -a '$actions --help'"""
        )

        command_blocks = [
            command_template.safe_substitute(
                command=op, actions=" ".join(list(actions.keys()))
            )
            for op, actions in self.ops.items()
        ]

        rendered = completion_template.safe_substitute(
            subcommands=" ".join(self.ops.keys()),
            command_items="\n".join(command_blocks),
        )

        return rendered

    def get_bash_completions(self):
        """
        Generates and returns bash shell completions based on the baked spec
        """
        completion_template = Template(
            """# This is a generated file!  Do not modify!
_linode_cli()
{
    local cur prev opts
    COMPREPLY=()
    cur="${COMP_WORDS[COMP_CWORD]}"
    prev="${COMP_WORDS[COMP_CWORD-1]}"

    case "${prev}" in
        linode-cli)
            COMPREPLY=( $(compgen -W "$actions --help" -- ${cur}) )
            return 0
            ;;
        $command_items
        *)
            ;;
    esac
}

complete -F _linode_cli linode-cli"""
        )

        command_template = Template(
            """$command)
            COMPREPLY=( $(compgen -W "$actions --help" -- ${cur}) )
            return 0
            ;;"""
        )

        command_blocks = [
            command_template.safe_substitute(
                command=op, actions=" ".join(list(actions.keys()))
            )
            for op, actions in self.ops.items()
        ]

        rendered = completion_template.safe_substitute(
            actions=" ".join(self.ops.keys()),
            command_items="\n        ".join(command_blocks),
        )

        return rendered

    def bake_completions(self):
        """
        Given a baked CLI, generates and saves a bash completion file
        """
        rendered = self.get_bash_completions()
        # save it off
        with open("linode-cli.sh", "w", encoding="utf-8") as f:
            print("Writing file...")
            f.write(rendered)

    def load_baked(self):
        """
        Loads a baked spec representation from a baked pickle
        """
        data_file = self._get_data_file()
        data_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), data_file)
        if os.path.exists(data_path):
            with open(data_path, "rb") as f:
                self.ops = pickle.load(f)
                if "_base_url" in self.ops:
                    self.base_url = self.ops["_base_url"]
                    del self.ops["_base_url"]
                if "_spec_version" in self.ops:
                    self.spec_version = self.ops["_spec_version"]
                    del self.ops["_spec_version"]
        else:
            print("No spec baked.  Please bake by calling this script as follows:")
            print("  python3 gen_cli.py bake /path/to/spec")
            self.ops = None  # this signals __init__.py to give up

    def _get_data_file(self):
        """
        Returns the name of the baked data file this program wants.  This is in
        part based on python version.
        """
        return f"data-{version_info[0]}"

    def print_request_debug_info(self, method, url, headers, body):
        """
        Prints debug info for an HTTP request
        """
        print(f"> {method.__name__.upper()} {url}", file=stderr)
        for k, v in headers.items():
            print(f"> {k}: {v}", file=stderr)
        print("> Body:", file=stderr)
        print(">  ", body or "", file=stderr)
        print("> ", file=stderr)

    def print_response_debug_info(self, response):
        """
        Prints debug info for a response from requests
        """
        # these come back as ints, convert to HTTP version
        http_version = response.raw.version / 10

        print(
            f"< HTTP/{http_version:.1f} {response.status_code} {response.reason}",
            file=stderr,
        )
        for k, v in response.headers.items():
            print(f"< {k}: {v}", file=stderr)
        print("< ", file=stderr)

    def do_request(
        self, operation, args, filter_header=None, skip_error_handling=False
    ):  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        """
        Makes a request to an operation's URL and returns the resulting JSON, or
        prints and error if a non-200 comes back
        """
        method = getattr(requests, operation.method)
        headers = {
            "Authorization": f"Bearer {self.config.get_token()}",
            "Content-Type": "application/json",
            "User-Agent": (
                f"linode-cli:{self.version} "
                f"python/{version_info[0]}.{version_info[1]}.{version_info[2]}"
            ),
        }

        parsed_args = operation.parse_args(args)

        url = operation.url.format(**vars(parsed_args))

        if operation.method == "get":
            url += f"?page={self.page}&page_size={self.page_size}"

        body = None
        if operation.method == "get":
            if filter_header is not None:
                # plugins can specify their own filters - use those by default
                headers["X-Filter"] = json.dumps(filter_header)
            else:
                # otherwise, get filters from the CLI call
                filters = vars(parsed_args)
                # remove URL parameters
                for p in operation.params:
                    if p.name in filters:
                        del filters[p.name]
                # remove empty filters
                filters = {k: v for k, v in filters.items() if v is not None}
                # apply filter, if any
                if filters:
                    headers["X-Filter"] = json.dumps(filters)
        else:
            if self.defaults:
                parsed_args = self.config.update(
                    parsed_args, operation.allowed_defaults
                )

            to_json = {k: v for k, v in vars(parsed_args).items() if v is not None}

            expanded_json = {}
            # expand paths
            for k, v in to_json.items():
                cur = expanded_json
                for part in k.split(".")[:-1]:
                    if part not in cur:
                        cur[part] = {}
                    cur = cur[part]
                cur[k.split(".")[-1]] = v

            body = json.dumps(expanded_json)

        if self.debug_request:
            self.print_request_debug_info(method, url, headers, body)

        result = method(url, headers=headers, data=body)

        if self.debug_request:
            self.print_response_debug_info(result)

        if not self.suppress_warnings:
            # check the major/minor version API reported against what we were built
            # with to see if an upgrade should be available
            api_version_higher = False

            if "X-Spec-Version" in result.headers:
                spec_version = result.headers.get("X-Spec-Version")

                try:
                    # Parse the spec versions from the API and local CLI.
                    StrictVersion(spec_version)
                    StrictVersion(self.spec_version)

                    # Get only the Major/Minor version of the API Spec and CLI Spec,
                    # ignore patch version differences
                    spec_major_minor_version = (
                        spec_version.split(".")[0] + "." + spec_version.split(".")[1]
                    )
                    current_major_minor_version = (
                        self.spec_version.split(".")[0]
                        + "."
                        + self.spec_version.split(".")[1]
                    )
                except ValueError:
                    # If versions are non-standard like, "DEVELOPMENT" use them and don't complain.
                    spec_major_minor_version = spec_version
                    current_major_minor_version = self.spec_version

                try:
                    if LooseVersion(spec_major_minor_version) > LooseVersion(
                        current_major_minor_version
                    ):
                        api_version_higher = True
                except:
                    # if this comparison or parsing failed, still process output
                    print(
                        f"Parsing failed when comparing local version {self.spec_version} with  "
                        f"server version {spec_version}.  If this problem persists, please open a "
                        "ticket with `linode-cli support ticket-create`",
                        file=stderr,
                    )

            if api_version_higher:
                # check to see if there is, in fact, a version to upgrade to.  If not, don't
                # suggest an upgrade (since there's no package anyway)
                new_version_exists = False

                try:
                    # do this all in a try block since it must _never_ prevent the CLI
                    # from showing command output
                    pypi_response = requests.get(
                        "https://pypi.org/pypi/linode-cli/json", timeout=1  # seconds
                    )

                    if pypi_response.status_code == 200:
                        # we got data back
                        pypi_version = pypi_response.json()["info"]["version"]

                        # no need to be fancy; these should always be valid versions
                        if LooseVersion(pypi_version) > LooseVersion(self.version):
                            new_version_exists = True
                except:
                    # I know, but if anything happens here the end user should still
                    # be able to see the command output
                    print(
                        "Unable to determine if a new linode-cli package is available "
                        "in pypi.  If this message persists, open a ticket or invoke "
                        "with --suppress-warnings",
                        file=stderr,
                    )

                if new_version_exists:
                    print(
                        f"The API responded with version {spec_version}, which is newer than "
                        f"the CLI's version of {self.spec_version}.  Please update the CLI to get "
                        "access to the newest features.  You can update with a "
                        f"simple `{PIP_CMD} install --upgrade linode-cli`",
                        file=stderr,
                    )

        if not 199 < result.status_code < 399 and not skip_error_handling:
            self._handle_error(result)

        return result

    def _handle_error(self, response):
        """
        Given an error message, properly displays the error to the user and exits.
        """
        print(f"Request failed: {response.status_code}", file=stderr)

        resp_json = response.json()

        if "errors" in resp_json:
            data = [
                [error.get("field") or "", error.get("reason")]
                for error in resp_json["errors"]
            ]
            self.output_handler.print(
                None, data, title="errors", to=stderr, columns=["field", "reason"]
            )
        sys.exit(1)

    @staticmethod
    def _flatten_url_path(tag):
        new_tag = tag.lower()
        new_tag = re.sub(r"[^a-z ]", "", new_tag).replace(" ", "-")
        return new_tag

    def handle_command(self, command, action, args):
        """
        Given a command, action, and remaining kwargs, finds and executes the
        action
        """

        if command not in self.ops:
            print(f"Command not found: {command}")
            sys.exit(1)

        operation = self.ops[command][action] if action in self.ops[command] else None

        if operation is None:
            # Find the matching alias
            for op in self.ops[command].values():
                if action in op.action_aliases:
                    operation = op
                    break

            # Fail if no matching alias was found
            if operation is None:
                print(f"No action {action} for command {command}")
                sys.exit(1)

        result = self.do_request(operation, args)

        operation.process_response_json(result.json(), self.output_handler)

        if (
            self.output_handler.mode == OutputMode.table
            and "pages" in result.json()
            and result.json()["pages"] > 1
        ):
            print(
                f"Page {result.json()['page']} of {result.json()['pages']}. "
                "Call with --page [PAGE] to load a different page."
            )

    def configure(self):
        """
        Reconfigure the application
        """
        self.config.configure()

    def call_operation(self, command, action, args=None, filters=None):
        """
        This function is used in plugins to retrieve the result of CLI operations
        in JSON format.  This uses the configured user of the CLI.

        :param filters: The X-Filter header to include in the request.  This overrides
                        whatever is passed into to command as filters.
        :type filters: dict
        """
        if args is None:
            args = []
        if command not in self.ops or action not in self.ops[command]:
            raise ValueError(f"Unknown command/action {command}/{action}")

        operation = self.ops[command][action]

        result = self.do_request(
            operation, args, filter_header=filters, skip_error_handling=True
        )

        return result.status_code, result.json()



#!/usr/local/bin/python3
"""
Argument parser for the linode CLI
"""

import argparse
import sys
import os
from importlib import import_module
from sys import argv, stderr, version_info

import pkg_resources
import requests
import yaml
from terminaltables import SingleTable

from linodecli import plugins

from .cli import CLI
from .configuration import ENV_TOKEN_NAME
from .operation import CLIArg, CLIOperation, URLParam
from .output import OutputMode
from .response import ModelAttr, ResponseModel

# this might not be installed at the time of building
try:
    VERSION = pkg_resources.require("linode-cli")[0].version
except:
    VERSION = "building"
BASE_URL = "https://api.linode.com/v4"


# if any of these arguments are given, we don't need to prompt for configuration
skip_config = any(c in argv for c in ["--skip-config", "--help", "--version"])

cli = CLI(VERSION, BASE_URL, skip_config=skip_config)


def warn_python2_eol():
    """
    Prints a warning the first time each day that the CLI is used under python2.
    This is included to help users upgrade to python3 before python2 support is
    dropped in the CLI.
    """
    if version_info.major < 3:
        print(
            "You are running the Linode CLI using Python 2, which reached its end of life on "
            "Jan 1st, 2020.  The Linode CLI will be dropping support for Python 2, and "
            "upgrading your installation is strongly encouraged so that you can continue "
            "to receive updates.\n\nFor information about upgrading your installation, see our "
            "official guide:\n\nhttps://www.linode.com/docs/guides/upgrade-to-linode-cli-python-3/",
            file=stderr,
        )


def main():  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    # TODO: major refactor function is too long
    """
    Handle incoming command arguments
    """
    parser = argparse.ArgumentParser("linode-cli", add_help=False)
    parser.add_argument(
        "command",
        metavar="COMMAND",
        nargs="?",
        type=str,
        help="The command to invoke in the CLI.",
    )
    parser.add_argument(
        "action",
        metavar="ACTION",
        nargs="?",
        type=str,
        help="The action to perform in this command.",
    )
    parser.add_argument(
        "--help",
        action="store_true",
        help="Display information about a command, action, or " "the CLI overall.",
    )
    parser.add_argument(
        "--text",
        action="store_true",
        help="Display text output with a delimiter (defaults to tabs).",
    )
    parser.add_argument(
        "--delimiter",
        metavar="DELIMITER",
        type=str,
        help="The delimiter when displaying raw output.",
    )
    parser.add_argument("--json", action="store_true", help="Display output as JSON")
    parser.add_argument(
        "--markdown", action="store_true", help="Display output in Markdown format."
    )
    parser.add_argument(
        "--pretty", action="store_true", help="If set, pretty-print JSON output"
    )
    parser.add_argument(
        "--no-headers",
        action="store_true",
        help="If set, does not display headers in output.",
    )
    parser.add_argument(
        "--page",
        metavar="PAGE",
        type=int,
        default=1,
        help="For listing actions, specifies the page to request",
    )
    parser.add_argument(
        "--page-size",
        metavar="PAGESIZE",
        type=int,
        default=100,
        help="For listing actions, specifies the number of items per page, "
        "accepts any value between 25 and 500",
    )
    parser.add_argument(
        "--all",
        action="store_true",
        help="If set, displays all possible columns instead of "
        "the default columns.  This may not work well on "
        "some terminals.",
    )
    parser.add_argument(
        "--format",
        metavar="FORMAT",
        type=str,
        help="The columns to display in output.  Provide a comma-"
        "separated list of column names.",
    )
    parser.add_argument(
        "--no-defaults",
        action="store_true",
        help="Suppress default values for arguments.  Default values "
        "are configured on initial setup or with linode-cli configure",
    )
    parser.add_argument(
        "--as-user",
        metavar="USERNAME",
        type=str,
        help="The username to execute this command as.  This user must "
        "be configured.",
    )
    parser.add_argument(
        "--suppress-warnings",
        action="store_true",
        help="Suppress warnings that are intended for human users. "
        "This is useful for scripting the CLI's behavior.",
    )
    parser.add_argument(
        "--version",
        "-v",
        action="store_true",
        help="Prints version information and exits.",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Enable verbose HTTP debug output"
    )

    parsed, args = parser.parse_known_args()

    # setup cli class
    if parsed.text:
        cli.output_handler.mode = OutputMode.delimited
    elif parsed.json:
        cli.output_handler.mode = OutputMode.json
        cli.output_handler.columns = "*"
    elif parsed.markdown:
        cli.output_handler.mode = OutputMode.markdown
    if parsed.delimiter:
        cli.output_handler.delimiter = parsed.delimiter
    if parsed.pretty:
        cli.output_handler.mode = OutputMode.json
        cli.output_handler.pretty_json = True
        cli.output_handler.columns = "*"
    if parsed.no_headers:
        cli.output_handler.headers = False
    if parsed.all:
        cli.output_handler.columns = "*"
    elif parsed.format:
        cli.output_handler.columns = parsed.format

    cli.defaults = not parsed.no_defaults
    cli.suppress_warnings = parsed.suppress_warnings
    cli.page = parsed.page
    cli.page_size = parsed.page_size
    cli.debug_request = parsed.debug

    if not cli.suppress_warnings:
        warn_python2_eol()

    if parsed.as_user and not skip_config:
        # if they are acting as a non-default user, set it up early
        cli.config.set_user(parsed.as_user)

    if parsed.version:
        if not parsed.command:
            # print version info and exit - but only if no command was given
            print(f"linode-cli {VERSION}")
            print(f"Built off spec version {cli.spec_version}")
            sys.exit(0)
        else:
            # something else might want to parse version
            # find where it was originally, as it was removed from args
            index = argv.index("--version") - 3  # executable command action
            args = args[:index] + ["--version"] + args[index:]

    # handle a bake - this is used to parse a spec and bake it as a pickle
    if parsed.command == "bake":
        if parsed.action is None:
            print("No spec provided, cannot bake")
            sys.exit(9)
        print("Baking...")
        spec_loc = parsed.action
        try:
            if os.path.exists(os.path.expanduser(spec_loc)):
                with open(os.path.expanduser(spec_loc), encoding="utf-8") as f:
                    spec = yaml.safe_load(f.read())
            else:  # try to GET it
                resp = requests.get(spec_loc, timeout=120)
                if resp.status_code == 200:
                    spec = yaml.safe_load(resp.content)
                else:
                    raise RuntimeError(f"Request failed to {spec_loc}")
        except Exception as e:
            print(f"Could not load spec: {e}")
            sys.exit(2)

        cli.bake(spec)
        print("Baking bash completions...")
        # this step would normally happen on laod
        if "_base_url" in cli.ops:
            del cli.ops["_base_url"]
        if "_spec_version" in cli.ops:
            del cli.ops["_spec_version"]
        # do the baking
        cli.bake_completions()
        print("Done.")
        sys.exit(0)
    elif cli.ops is None:
        # if not spec was found and we weren't baking, we're doomed
        sys.exit(3)

    if parsed.command == "register-plugin":
        # this is how the CLI discovers third-party plugins - the user registers
        # them!  Registering a plugin gets it set up for all CLI users.
        if parsed.action is None:
            print("register-plugin requires a module name!")
            sys.exit(9)
        module = parsed.action

        # attempt to import the module to prove it is installed and exists
        try:
            plugin = import_module(module)
        except ImportError:
            print(f"Module {module} not installed")
            sys.exit(10)

        # get the plugin name
        try:
            plugin_name = plugin.PLUGIN_NAME
        except AttributeError:
            print(f"{module} is not a valid Linode CLI plugin - missing PLUGIN_NAME")
            sys.exit(11)

        # prove it's callable
        try:
            call_func = plugin.call
            del call_func
        except AttributeError:
            print(f"{module} is not a valid Linode CLI plugin - missing call")
            sys.exit(11)

        reregistering = False
        # check for naming conflicts
        if plugin_name in cli.ops:
            print("Plugin name conflicts with CLI operation - registration failed.")
            sys.exit(12)
        elif plugin_name in plugins.available_local:
            # conflicts with an internal plugin - can't do that
            print(
                "Plugin name conflicts with internal CLI plugin - registration failed."
            )
            sys.exit(13)
        elif plugin_name in plugins.available(cli.config):
            # this isn't an internal plugin, so warn that we're re-registering it
            print(f"WARNING: Plugin {plugin_name} is already registered.")
            print("")
            answer = input(f"Allow re-registration of {plugin_name}? [y/N] ")

            if not answer or answer not in "yY":
                print("Registration aborted.")
                sys.exit(0)

            reregistering = True

        # looks good - register it
        already_registered = []
        if cli.config.config.has_option("DEFAULT", "registered-plugins"):
            already_registered = cli.config.config.get(
                "DEFAULT", "registered-plugins"
            ).split(",")

        if reregistering:
            already_registered.remove(plugin_name)
            cli.config.config.remove_option("DEFAULT", f"plugin-name-{plugin_name}")

        already_registered.append(plugin_name)
        cli.config.config.set(
            "DEFAULT", "registered-plugins", ",".join(already_registered)
        )
        cli.config.config.set("DEFAULT", f"plugin-name-{plugin_name}", module)
        cli.config.write_config()

        print("Plugin registered successfully!")
        print("Invoke this plugin by running the following:")
        print(f"  linode-cli {plugin_name}")
        sys.exit(0)

    if parsed.command == "remove-plugin":
        if parsed.action is None:
            print("remove-plugin requires a plugin name to remove!")
            sys.exit(9)

        # is this plugin registered?
        plugin_name = parsed.action
        if plugin_name in plugins.available_local:
            # can't remove first-party plugins
            print(f"{plugin_name} is bundled with the CLI and cannot be removed")
            sys.exit(13)
        elif plugin_name not in plugins.available(cli.config):
            print(f"{plugin_name} is not a registered plugin")
            sys.exit(14)

        # do the removal
        current_plugins = cli.config.config.get("DEFAULT", "registered-plugins").split(
            ","
        )
        current_plugins.remove(plugin_name)
        cli.config.config.set(
            "DEFAULT", "registered-plugins", ",".join(current_plugins)
        )

        if cli.config.config.has_option("DEFAULT", f"plugin-name-{plugin_name}"):
            # if the config if malformed, don't blow up
            cli.config.config.remove_option("DEFAULT", f"plugin-name-{plugin_name}")

        cli.config.write_config()

        print(f"Plugin {plugin_name} removed")
        sys.exit(0)

    if parsed.command == "completion":
        if parsed.help or not parsed.action:
            print("linode-cli completion [SHELL]")
            print()
            print(
                "Prints shell completions for the requested shell to stdout. "
                "Currently, only completions for bash and fish are available."
            )
            sys.exit(0)

        completions = ""

        if parsed.action == "bash":
            completions = cli.get_bash_completions()
        elif parsed.action == "fish":
            completions = cli.get_fish_completions()
        else:
            print(
                "Completions are only available for bash and fish at this time.  To retrieve "
                "these, please invoke as `linode-cli completion bash` "
                "or `linode-cli completion fish`."
            )
            sys.exit(1)
        print(completions)
        sys.exit(0)

    # handle a help for the CLI
    if parsed.command is None or (parsed.command is None and parsed.help):
        parser.print_help()

        # commands to manage CLI users (don't call out to API)
        print()
        print("CLI user management commands:")
        um_commands = [["configure", "set-user", "show-users"], ["remove-user"]]
        table = SingleTable(um_commands)
        table.inner_heading_row_border = False
        print(table.table)

        # commands to manage plugins (don't call out to API)
        print()
        print("CLI Plugin management commands:")
        pm_commands = [["register-plugin", "remove-plugin"]]
        table = SingleTable(pm_commands)
        table.inner_heading_row_border = False
        print(table.table)

        # other CLI commands
        print()
        print("Other CLI commands:")
        other_commands = [["completion"]]
        table = SingleTable(other_commands)
        table.inner_heading_row_border = False
        print(table.table)

        # commands generated from the spec (call the API directly)
        print()
        print("Available commands:")

        content = list(sorted(cli.ops.keys()))
        proc = []
        for i in range(0, len(content), 3):
            proc.append(content[i : i + 3])
        if content[i + 3 :]:
            proc.append(content[i + 3 :])

        table = SingleTable(proc)
        table.inner_heading_row_border = False
        print(table.table)

        # plugins registered to the CLI (do arbitrary things)
        if plugins.available(cli.config):
            # only show this if there are any available plugins
            print("Available plugins:")

            plugin_content = list(plugins.available(cli.config))
            plugin_proc = []

            for i in range(0, len(plugin_content), 3):
                plugin_proc.append(plugin_content[i : i + 3])
            if plugin_content[i + 3 :]:
                plugin_proc.append(plugin_content[i + 3 :])

            plugin_table = SingleTable(plugin_proc)
            plugin_table.inner_heading_row_border = False

            print(plugin_table.table)

        print()
        print("To reconfigure, call `linode-cli configure`")
        print("For comprehensive documentation, visit https://www.linode.com/docs/api/")
        sys.exit(0)

    # configure
    if parsed.command == "configure":
        if parsed.help:
            print("linode-cli configure")
            print()
            print("Configured the Linode CLI.  This command can be used to change")
            print("defaults selected for the current user, or to configure additional")
            print("users.")
            sys.exit(0)
        else:
            cli.configure()
            sys.exit(0)

    # block of commands for user-focused operations
    if parsed.command == "set-user":
        if parsed.help or not parsed.action:
            print("linode-cli set-user [USER]")
            print()
            print("Sets the active user for the CLI out of users you have configured.")
            print("To configure a new user, see `linode-cli configure`")
            sys.exit(0)
        else:
            cli.config.set_default_user(parsed.action)
            sys.exit(0)

    if parsed.command == "show-users":
        if parsed.help:
            print("linode-cli show-users")
            print()
            print("Lists configured users.  Configured users can be set as the")
            print("active user (used for all commands going forward) with the")
            print("`set-user` command, or used for a single command with the")
            print(
                "`--as-user` flag.  New users can be added with `linode-cli configure`."
            )
            print("The user that is currently active is indicated with a `*`")
            sys.exit(0)
        else:
            cli.config.print_users()
            sys.exit(0)

    if parsed.command == "remove-user":
        if parsed.help or not parsed.action:
            print("linode-cli remove-user [USER]")
            print()
            print("Removes a user the CLI was configured with.  This does not change")
            print("your Linode account, only this CLI installation.  Once removed,")
            print("the user may not be set as active or used for commands unless")
            print("configured again.")
            sys.exit(0)
        else:
            cli.config.remove_user(parsed.action)
            sys.exit(0)

    # special command to bake shell completion script
    if parsed.command == "bake-bash":
        cli.bake_completions()

    # check for plugin invocation
    if parsed.command not in cli.ops and parsed.command in plugins.available(
        cli.config
    ):
        context = plugins.PluginContext(cli.config.get_token(), cli)

        # reconstruct arguments to send to the plugin
        plugin_args = argv[1:]  # don't include the program name
        plugin_args.remove(parsed.command)  # don't include the plugin name tho

        plugins.invoke(parsed.command, plugin_args, context)
        sys.exit(0)

    if parsed.command not in cli.ops and parsed.command not in plugins.available(
        cli.config
    ):
        # unknown commands
        print(f"Unrecognized command {parsed.command}")

    # handle a help for a command - either --help or no action triggers this
    if parsed.command is not None and parsed.action is None:
        if parsed.command in cli.ops:
            actions = cli.ops[parsed.command]
            print(f"linode-cli {parsed.command} [ACTION]")
            print()
            print("Available actions: ")

            content = [
                [", ".join([action, *op.action_aliases]), op.summary]
                for action, op in actions.items()
            ]

            header = ["action", "summary"]
            table = SingleTable([header] + content)
            print(table.table)
            sys.exit(0)

    # handle a help for an action
    if parsed.command is not None and parsed.action is not None and parsed.help:
        if parsed.command in cli.ops and parsed.action in cli.ops[parsed.command]:
            operation = cli.ops[parsed.command][parsed.action]
            print(f"linode-cli {parsed.command} {parsed.action}", end="")
            for param in operation.params:
                # clean up parameter names - we add an '_' at the end of them
                # during baking if it conflicts with the name of an argument.
                # Remove the trailing underscores on output (they're not
                # important to the end user).
                pname = param.name.upper()
                if pname[-1] == "_":
                    pname = pname[:-1]
                print(f" [{pname}]", end="")
            print()
            print(operation.summary)
            if operation.docs_url:
                print(f"API Documentation: {operation.docs_url}")
            print()
            if operation.args:
                print("Arguments:")
                for arg in sorted(operation.args, key=lambda s: not s.required):
                    is_required = (
                        "(required) "
                        if operation.method in {"post", "put"} and arg.required
                        else ""
                    )
                    print(f"  --{arg.path}: {is_required}{arg.description}")
            elif operation.method == "get" and parsed.action == "list":
                filterable_attrs = [
                    attr for attr in operation.response_model.attrs if attr.filterable
                ]

                if filterable_attrs:
                    print("You may filter results with:")
                    for attr in filterable_attrs:
                        print(f"  --{attr.name}")
        sys.exit(0)

    if parsed.command is not None and parsed.action is not None:
        cli.handle_command(parsed.command, parsed.action, args)



#!/usr/bin/env python3
import subprocess
from os import path

from setuptools import setup

here = path.abspath(path.dirname(__file__))


# get the long description from the README.rst
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()


def get_baked_files():
    """
    A helper to retrieve the baked files included with this package.  This is
    to assist with building from source, where baked files may not be present
    on a fresh clone
    """
    data_files = []

    if path.isfile("linode-cli.sh"):
        data_files.append(("/etc/bash_completion.d", ["linode-cli.sh"]))

    return data_files


def get_version():
    """
    Uses the version file to calculate this package's version
    """
    return subprocess.check_output(["./version"]).decode("utf-8").rstrip()


def get_baked_version():
    """
    Attempts to read the version from the baked_version file
    """
    with open("./baked_version", "r", encoding="utf-8") as f:
        result = f.read()

    return result


def bake_version(v):
    """
    Writes the given version to the baked_version file
    """
    with open("./baked_version", "w", encoding="utf-8") as f:
        f.write(v)


# If there's already a baked version, use it.
# This is useful for installing from an SDist.
# NOTE: baked_version is deleted when running `make clean`.
if path.isfile("baked_version"):
    version = get_baked_version()
else:
    # Otherwise, retrieve and bake the version as normal
    version = get_version()
    bake_version(version)

setup(
    name="linode-cli-testrelease",
    version=version,
    description="CLI for the Linode API",
    long_description=long_description,
    author="Linode",
    author_email="developers@linode.com",
    url="https://www.linode.com/docs/api/",
    packages=[
        "linodecli",
        "linodecli.plugins",
    ],
    license="BSD 3-Clause License",
    install_requires=[
        "terminaltables",
        "requests",
        "PyYAML",
        "future; python_version <= '3.0.0'",
    ],
    extras_require={
        ":python_version<'3.4'": ["enum34"],
    },
    entry_points={
        "console_scripts": [
            "linode-cli = linodecli:main",
        ]
    },
    data_files=get_baked_files(),
    python_requires=">=3.6",
    include_package_data=True,
)



"""
The ssh plugin allows sshing into Linodes by label or ID

Invoke as follows::

   linode-cli ssh [USERNAME@]LINODE_LABEL [SSH ARGS..]

   LINODE_LABEL - the label of the Linode to ssh into
   USERNAME - the user to ssh into the Linode as.  Defaults to the current user
"""
import argparse
import subprocess
import sys
from sys import platform


def call(args, context):  # pylint: disable=too-many-branches
    """
    Invokes this plugin
    """
    if platform == "win32":
        print(
            "This plugin is not currently supported in Windows.  For more "
            "information or to suggest a fix, please visit "
            "https://github.com/linode/linode-cli"
        )
        sys.exit(1)

    parser = argparse.ArgumentParser("linode-cli ssh", add_help=True)
    parser.add_argument(
        "label",
        metavar="[USERNAME@]LABEL",
        nargs="?",
        type=str,
        help="The label of the Linode to SSH into, optionally with "
        "a username before it in USERNAME@LABEL format.  If no "
        "username is given, defaults to the current user.",
    )
    parser.add_argument(
        "-6",
        action="store_true",
        help="If given, uses the Linode's SLAAC address for SSH.",
    )

    parsed, args = parser.parse_known_args(args)

    if not parsed.label:
        parser.print_help()
        sys.exit(0)

    label = parsed.label
    username = None
    if "@" in parsed.label:
        username, label = parsed.label.split("@", 1)

    result, potential_matches = context.client.call_operation(
        "linodes", "list", filters={"label": {"+contains": label}}
    )

    if result != 200:
        print(f"Could not retrieve Linode: {result} error")
        sys.exit(2)

    potential_matches = potential_matches["data"]
    exact_match = None

    # see if we got a match
    for match in potential_matches:
        if match["label"] == label:
            exact_match = match
            break

    if exact_match is None:
        # no match - stop
        print(f"No Linode found for label {label}")

        if potential_matches:
            print("Did you mean: ")
            print("\n".join([f" {p['label']}" for p in potential_matches]))
        sys.exit(1)

    if exact_match["status"] != "running":
        print(
            f"{label} is not running (status is {exact_match['status']}); operation aborted."
        )
        sys.exit(2)

    # find a public IP Address to use
    public_ip = None

    if getattr(
        parsed, "6"
    ):  # this is necessary since the name isn't a valid python variable name
        public_ip = exact_match["ipv6"].split("/")[0]
    else:
        for ip in exact_match["ipv4"]:
            if not ip.startswith("192.168"):
                public_ip = ip  # TODO - this uses the "first" IP Address
                break

    address = public_ip
    if username:
        address = username + "@" + address

    # do it
    code = 0
    try:
        # tack the remaining unparsed args onto the end - those are args for ssh
        subprocess.check_call(["ssh", address] + args)
    except subprocess.CalledProcessError as e:
        # ssh exited with non-zero status code
        code = e.returncode

    # exit with the same code as ssh
    sys.exit(code)



"""
Plugin for CLI for editing firewalls
"""

import argparse
import json
import re
import termios
import sys
from typing import Callable
from ipaddress import ip_address, IPv4Address

from terminaltables import PorcelainTable

BOLD = "\033[1m"
NOT_BOLD = "\033[0m"

# plugin-scoped variables
# setup terminal states
orig_term = termios.tcgetattr(sys.stdin.fileno())
new_term = termios.tcgetattr(sys.stdin.fileno())
# new terminal becomes unbuffered
new_term[3] = new_term[3] & ~termios.ICANON & ~termios.ECHO


class InputValidation:
    """
    Validating the input of firewalls
    """

    @staticmethod
    def input(input_text: str, validator: Callable[[str], None]):
        """
        Handle input from user
        """
        while True:
            value = input(input_text)
            try:
                validator(value)
            except ValueError as err:
                print(f"Invalid Input: {'; '.join(err.args)}")
                continue

            return value

    @staticmethod
    def input_io(rules):
        """
        Handle input option between Inbound and Outbound
        """
        raw_terminal()

        print("[I]nbound or [O]utbound?")
        segment = " "
        while segment not in "io":
            segment = sys.stdin.read(1).lower()

        change = rules["inbound"]
        if segment == "o":
            change = rules["outbound"]

        revert_terminal()

        return change

    @staticmethod
    def optional(validator):
        """
        Handle optional value
        """

        def callback(value):
            if value.strip() == "":
                return

            validator(value)

        return callback

    @staticmethod
    def int():
        """
        Check if value is numeric
        """

        def callback(value):
            if not value.isnumeric():
                raise ValueError(f"Expected an integer, got {value}")

        return callback

    @staticmethod
    def index_of(ref_list, allow_append=False):
        """
        Validate index
        """

        def callback(value):
            if not value.isnumeric():
                raise ValueError(f"Expected an integer, got {value}")

            value_int = int(value)

            if value_int < 0 or value_int >= len(ref_list) + (1 if allow_append else 0):
                raise ValueError(f"Invalid index {value_int}")

        return callback

    @staticmethod
    def regex(pattern, error_msg):
        """
        Regex callback
        """

        def callback(value):
            pattern_compiled = re.compile(pattern)

            if not re.fullmatch(pattern_compiled, value):
                raise ValueError(error_msg)

        return callback

    @staticmethod
    def one_of(valid_choices):
        """
        Validated choice is one of choices
        """

        def callback(value):
            if value.lower() not in [choice.lower() for choice in valid_choices]:
                raise ValueError(
                    f"Invalid choice: {value}; must be one "
                    f"of {', '.join(valid_choices)}"
                )

        return callback

    @staticmethod
    def ip_list():
        """
        Callback for IP List
        """

        def callback(value):
            for ip in value.split(","):
                ip = ip.strip()

                # IP ranges are required
                ip_parts = ip.split("/")

                if len(ip_parts) != 2:
                    raise ValueError(
                        f"Invalid IP: {ip}; IPs must be in IP/mask format."
                    )

                if not ip_parts[1].isnumeric():
                    raise ValueError(f"Invalid IP: {ip}; IP masks must be numeric")

                try:
                    ip_address(ip_parts[0])
                except ValueError as e:
                    raise ValueError(f"Invalid IP: {ip}") from e

        return callback


def raw_terminal():
    """
    Sets the terminal to new_term, in raw mode
    """
    termios.tcsetattr(sys.stdin.fileno(), termios.TCSAFLUSH, new_term)


def revert_terminal():
    """
    Sets the terminal to orig_term, as we found it
    """
    # and before we leave, reset to original terminal
    termios.tcsetattr(sys.stdin.fileno(), termios.TCSAFLUSH, orig_term)


##
# Helper functions
##


def _get_firewall(firewall_id, client):
    """
    Returns the firewall object with the given ID
    """
    code, firewall = client.call_operation("firewalls", "view", args=[firewall_id])

    if code != 200:
        print(f"Error retrieving firewall: {code}")
        sys.exit(1)

    code, rules = client.call_operation("firewalls", "rules-list", args=[firewall_id])

    if code != 200:
        print(f"Error retrieving firewall rules: {code}")
        sys.exit(2)

    return firewall, rules


##
# Drawing functions
##


def redraw(firewall, rules):
    """
    Clears the screen and redraws the firewall header
    """
    # clear the whole screen
    print("\033[H\033[J", end="")

    # print the title bar
    print(
        "{bold}Firewall:{nobold} {label}\t{bold}Status:{nobold} {status}".format(
            bold=BOLD,
            nobold=NOT_BOLD,
            label=firewall["label"],
            status=firewall["status"],
        )
    )
    print(
        f"{BOLD}Inbound Policy:{NOT_BOLD} {rules['inbound_policy']}\t"
        f"{BOLD}Outbound Policy:{NOT_BOLD} {rules['outbound_policy']}"
    )


def print_rules_table(rules):
    """
    Prints a table of rules to the screen with a given index offset
    """
    if len(rules) < 1:
        print("None")
        return

    header = ["#", "label", "protocol", "action", "ports", "addresses"]
    rows = []

    for ind, cur in enumerate(rules):
        addrs = []
        if "ipv4" in cur["addresses"]:
            addrs.extend(cur["addresses"]["ipv4"])

        if "ipv6" in cur["addresses"]:
            addrs.extend(cur["addresses"]["ipv6"])

        rows.append(
            [
                ind,
                cur["label"] if "label" in cur else "",
                cur["protocol"],
                cur["action"],
                cur["ports"] if "ports" in cur else "",
                ", ".join(addrs),
            ]
        )

    tab = PorcelainTable([header] + rows)
    tab.inner_heading_row_border = True
    print(tab.table)


def draw_rules(rules):
    """
    Draws the current rules
    """
    # clear the rules and below
    print("\033[3H\033[J", end="")

    print()
    print(f"{BOLD}Inbound Rules:{NOT_BOLD}")
    print_rules_table(rules["inbound"])
    print()
    print(f"{BOLD}Outbound Rules:{NOT_BOLD}")
    print_rules_table(rules["outbound"])


##
# Custom Exceptions
##


class StopDontSave(KeyboardInterrupt):
    """
    Excpetion that indicates that we shouldn't save changes; overrides KeyboardInterrupt
    so that ^C inherits this behavior
    """


class StopSave(Exception):
    """
    Custom exception class raised to break the main loop, saving when we're done
    """


##
# ACTIONS
##
# The signature of these is:
#
#    def action_name(rules: dict) -> bool:
#
# Where the incoming dict is the current state of the firewall rules (which will be
# modified), and the return value if whether or not we need to redraw the whole screen
##


def save_quit(rules):
    """
    Handle quitting with saving
    """
    raise StopSave()


def quit(rules):  # pylint: disable=redefined-builtin
    """
    Soft handle quitting without saving
    """
    raise StopDontSave()


def add_rule(rules):
    """
    Adds a new firewall rule
    """
    change = InputValidation.input_io(rules)

    ind = 0
    if len(change) > 0:
        ind_str = InputValidation.input(
            "Index (Optional): ",
            InputValidation.optional(
                InputValidation.index_of(change, allow_append=True)
            ),
        ).strip()

        if ind_str == "":
            ind = len(change)
        else:
            ind = int(ind_str)

    label = InputValidation.input(
        "Rule Label (Optional): ",
        InputValidation.optional(
            InputValidation.regex(
                "^[a-zA-Z0-9\\-\\_\\.]{3,32}$",
                "Label must include only ASCII letters, numbers, "
                + "underscores, periods, and dashes.",
            )
        ),
    ).strip()

    protocol = InputValidation.input(
        "Protocol (TCP/UDP/ICMP/IPENCAP): ",
        InputValidation.one_of({"TCP", "UDP", "ICMP", "IPENCAP"}),
    )
    protocol = protocol.upper()

    action = InputValidation.input(
        "Action (ACCEPT/DROP): ", InputValidation.one_of({"ACCEPT", "DROP"})
    )
    action = action.upper()

    # Ports are now allowed for ICMP and IPENCAP protocols
    port = None

    if protocol not in ["ICMP", "IPENCAP"]:
        port = InputValidation.input(
            "Ports: ",
            InputValidation.regex(
                "^[0-9\\,\\- ]*$",
                "Input may be a single port, a range of ports, or a "
                "comma-separated list of single ports and port ranges.",
            ),
        )

    addresses = InputValidation.input(
        "Addresses (comma separated): ",
        InputValidation.ip_list(),
    )

    addresses_ipv4 = []
    addresses_ipv6 = []

    for ip in addresses.split(","):
        ip = ip.strip()

        if (
            type(ip_address(ip.split("/")[0])) is IPv4Address
        ):  # pylint: disable=unidiomatic-typecheck
            addresses_ipv4.append(ip)
        else:
            addresses_ipv6.append(ip)

    # make the new rule
    new_rule = {
        "protocol": protocol,
        "addresses": {},
        "action": action,
    }

    # Ports are rejected for ICMP and IPENCAP protocols
    if port is not None:
        new_rule["ports"] = port

    # The API will reject empty lists, let's make sure
    # they're only included if necessary
    if len(addresses_ipv4) > 0:
        new_rule["addresses"]["ipv4"] = addresses_ipv4

    if len(addresses_ipv6) > 0:
        new_rule["addresses"]["ipv6"] = addresses_ipv6

    if label:
        # this field must be omitted unless it is set
        new_rule["label"] = label

    change.insert(ind, new_rule)

    return False


def remove_rule(rules):
    """
    Removes a rule (by index)
    """
    change = InputValidation.input_io(rules)

    if len(change) < 1:
        print("No entires to remove")
        return False

    ind_str = InputValidation.input(
        "Index to remove: ", InputValidation.optional(InputValidation.index_of(change))
    ).strip()

    # No changes to be made
    if ind_str == "":
        return False

    del change[int(ind_str)]

    return False


def swap_rules(rules):
    """
    Swap index rules
    """
    revert_terminal()

    change = InputValidation.input_io(rules)

    a_str = InputValidation.input(
        "Swap index: ", InputValidation.optional(InputValidation.index_of(change))
    ).strip()

    if a_str == "":
        return False

    b_str = InputValidation.input(
        "With index: ", InputValidation.optional(InputValidation.index_of(change))
    ).strip()

    if b_str == "":
        return False

    a, b = int(a_str), int(b_str)

    change[a], change[b] = change[b], change[a]

    return False


def toggle_policy(policy_key):
    """
    Callback for toggling policy
    """

    def callback(rules):
        rules[policy_key] = "DROP" if rules[policy_key] == "ACCEPT" else "ACCEPT"

        return True

    return callback


ACTION_MAP = {
    "w": save_quit,
    "q": quit,
    "a": add_rule,
    "r": remove_rule,
    "s": swap_rules,
    "i": toggle_policy("inbound_policy"),
    "o": toggle_policy("outbound_policy"),
}


##
# Logic
##


def get_action():
    """
    Prints the action prompt and returns the action to run
    """
    print()
    print("Rules: [A]dd, [R]emove, or [S]wap")
    print("Policy: Toggle [I]nbound or [O]utbound")
    print("[W]rite settings or [Q]uit")

    selection = " "

    while selection.lower() not in ACTION_MAP:
        selection = sys.stdin.read(1)

    return ACTION_MAP[selection.lower()]


def mainloop(firewall, rules):
    """
    Accepts a firewall and the current ruleset and loops until told to be done
    allowing the rules to be modified.
    """
    full_redraw = True
    save = False

    while True:
        try:
            if full_redraw:
                redraw(firewall, rules)
                full_redraw = False

            draw_rules(rules)

            raw_terminal()
            action = get_action()

            full_redraw = action(rules)
        except StopSave:
            save = True
            break
        except StopDontSave:
            break

    revert_terminal()

    return save


def call(args, context):
    """
    Invokes the Interactive Firewall Plugin
    """
    parser = argparse.ArgumentParser("firewall-editor", add_help=True)
    parser.add_argument("firewall_id", help="The ID of the firewall to edit.")
    parsed = parser.parse_args(args)

    # fetch the info and start going
    firewall, rules = _get_firewall(parsed.firewall_id, context.client)

    while True:
        save = mainloop(firewall, rules)

        # then update the rules
        if save:
            print("Saving Firewall configuration..")
            code, errors = context.client.call_operation(
                "firewalls",
                "rules-update",
                args=[
                    parsed.firewall_id,
                    "--inbound",
                    json.dumps(rules["inbound"]),
                    "--outbound",
                    json.dumps(rules["outbound"]),
                    "--inbound_policy",
                    rules["inbound_policy"],
                    "--outbound_policy",
                    rules["outbound_policy"],
                ],
            )
            if code == 200:
                print("Rules updated successfully!")
                break
            print(f"Error editing rules: {code}: {errors}")
            # block to see the error, then re-enter the editor
            sys.stdin.read(1)
        else:
            print("Aborted.")
            break



"""
Handles formatting the output of commands used in Linode CLI
"""
import json
from enum import Enum
from sys import stdout

from terminaltables import SingleTable


class OutputMode(Enum):
    """
    Enum for output modes
    """

    table = 1
    delimited = 2
    json = 3
    markdown = 4


class OutputHandler:  # pylint: disable=too-few-public-methods
    """
    Handles formatting the output of commands used in Linode CLI
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        mode=OutputMode.table,
        delimiter="\t",
        headers=True,
        pretty_json=False,
        columns=None,
    ):
        self.mode = mode
        self.delimiter = delimiter
        self.pretty_json = pretty_json
        self.headers = headers
        self.columns = columns

    def print(
        self, response_model, data, title=None, to=stdout, columns=None
    ):  # pylint: disable=too-many-arguments
        """
        :param response_model: The Model corresponding to this response
        :type response_model: ResponseModel
        :param data: The data to display
        :type data: list[str] or list[dict]
        :param title: The title to display on a table
        :type title: str or None
        :param to: Where to print output to
        :type to: stdout, stderr or file
        :param columns: The columns to display
        :type columns: list[str]
        """
        if columns is None:
            columns = self._get_columns(response_model)
            header = [c.column_name for c in columns]
        else:
            header = columns

        if self.mode == OutputMode.table:
            self._table_output(header, data, columns, title, to)
        elif self.mode == OutputMode.delimited:
            self._delimited_output(header, data, columns, to)
        elif self.mode == OutputMode.json:
            self._json_output(header, data, to)
        elif self.mode == OutputMode.markdown:
            self._markdown_output(header, data, columns)

    def _get_columns(self, response_model):
        """
        Based on the configured columns, returns columns from a response model
        """
        if self.columns is None:
            columns = [
                attr
                for attr in sorted(response_model.attrs, key=lambda c: c.display)
                if attr.display
            ]
        elif self.columns == "*":
            columns = list(response_model.attrs)
        else:
            columns = []
            for col in self.columns.split(","):
                for attr in response_model.attrs:
                    if attr.column_name == col:
                        response_model.attrs.remove(attr)
                        columns.append(attr)
                        continue

        if not columns:
            # either they selected nothing, or the model wasn't setup for CLI
            # display - either way, display everything
            columns = response_model.attrs

        return columns

    def _table_output(
        self, header, data, columns, title, to
    ):  # pylint: disable=too-many-arguments
        """
        Pretty-prints data in a table
        """
        content = []

        if isinstance(columns[0], str):
            content = data
        else:
            for model in data:
                content.append([attr.render_value(model) for attr in columns])

        if self.headers:
            content = [header] + content

        tab = SingleTable(content)

        if title is not None:
            tab.title = title

        if not self.headers:
            tab.inner_heading_row_border = False

        print(tab.table, file=to)

    def _delimited_output(self, header, data, columns, to):
        """
        Prints data in delimited format with the given delimiter
        """
        content = []

        if isinstance(columns[0], str):
            content = data
        else:
            for model in data:
                content.append([attr.get_string(model) for attr in columns])

        if self.headers:
            content = [header] + content

        for row in content:
            print(self.delimiter.join(row), file=to)

    def _json_output(self, header, data, to):
        """
        Prints data in JSON format
        """
        content = []
        if len(data) and isinstance(data[0], dict):  # we got delimited json in
            # parse down to the value we display
            for row in data:
                content.append(self._select_json_elements(header, row))
        else:  # this is a list
            for row in data:
                content.append(dict(zip(header, row)))

        print(
            json.dumps(
                content,
                indent=2 if self.pretty_json else None,
                sort_keys=self.pretty_json,
            ),
            file=to,
        )

    def _select_json_elements(self, keys, json_res):
        """
        Returns a dict filtered down to include only the selected keys.  Walks
        paths to handle nested dicts
        """
        ret = {}
        for k, v in json_res.items():
            if k in keys:
                ret[k] = v
            elif isinstance(v, dict):
                v = self._select_json_elements(keys, v)
                if v:
                    ret[k] = v
        return ret

    def _markdown_output(self, header, data, columns):
        """
        Pretty-prints data in a Markdown-formatted table.  This uses github's
        flavor of Markdown
        """
        content = []

        if isinstance(columns[0], str):
            content = data
        else:
            for model in data:
                content.append(
                    [attr.render_value(model, colorize=False) for attr in columns]
                )

        if header:
            print("| " + " | ".join([str(c) for c in header]) + " |")
            print("|---" * len(header) + "|")

        for row in content:
            print("| " + " | ".join([str(c) for c in row]) + " |")
